import json
import os
import socket
from platform import uname
from shutil import which
from typing import List, Tuple

import click
import docker
from kubernetes import client as k8sclient

from openapi_client.models import ResponseKernelClusterInfo
from vessl.cli._base import VesslGroup, vessl_argument, vessl_option
from vessl.cli._util import (
    Endpoint,
    format_string,
    generic_prompter,
    print_data,
    print_debug,
    print_error,
    print_error_result,
    print_info,
    print_prompt_style,
    print_success,
    print_success_result,
    print_table,
    print_warning,
    prompt_choices,
    prompt_confirm,
)
from vessl.cli.organization import organization_name_option
from vessl.kernel_cluster import (
    CreateClusterParam,
    _acquire_kubeconfig,
    _check_existing_vessl_installation,
    _get_running_k0s_container,
    _load_cluster_kubeconfig,
    _load_kubeconfig_to_config,
    _start_k0s_container,
    _wait_for_k0s_ready,
    add_and_update_helm_repo,
    create_cluster,
    delete_cluster,
    list_cluster_nodes,
    list_clusters,
    read_cluster,
    rename_cluster,
)
from vessl.kernel_cluster_install_command_builder import ClusterInstallCommandBuilder
from vessl.organization import _get_organization_name
from vessl.util.config import VesslConfigLoader
from vessl.util.constant import CLUSTER_MODE_SINGLE, VESSL_HELM_REPO
from vessl.util.exception import (
    ACCESS_TOKEN_NOT_FOUND_ERROR_CODE,
    InvalidKernelClusterError,
    VesslApiException,
)


def cluster_name_prompter(ctx: click.Context, param: click.Parameter, value: str) -> str:
    clusters = list_clusters()
    return prompt_choices("Cluster", [x.name for x in clusters])


def custom_cluster_name_prompter(ctx: click.Context, param: click.Parameter, value: str) -> str:
    clusters = list_clusters()
    clusters = [x for x in clusters if not x.is_savvihub_managed]
    return prompt_choices("Cluster", [x.name for x in clusters])


@click.command(name="cluster", cls=VesslGroup)
def cli():
    pass


@cli.vessl_command()
@vessl_option(
    "--name",
    type=click.STRING,
    required=True,
    help="Name of the cluster. (e.g. 'seoul-cluster')",
    prompt="Enter cluster name",
)
@vessl_option(
    "--kubernetes-namespace",
    type=click.STRING,
    default="vessl",
    help="Kubernetes namespace to install VESSL agent (default: vessl)",
    prompt="Enter Kubernetes namespace to install VESSL agent",
    prompt_required=True,
)
@vessl_option(
    "--provider",
    type=click.Choice(["aws", "gcp", "on-premise"], case_sensitive=False),
    default="on-premise",
    help="Cluster provider; aws|gcp|on-premise (default: on-premise)",
    prompt="Select your cluster provider",
    prompt_required=True,
)
@vessl_option(
    "--extra-helm-values",
    type=click.STRING,
    multiple=True,
    default=[],
    help="Extra Helm values to pass when installing VESSL agent.",
)
@organization_name_option
def create(
    name: str,
    kubernetes_namespace: str,
    provider: str,
    extra_helm_values: List[str],
    organization_name: str = None,
):
    ########################################
    # validate inputs
    ########################################
    if name is None or len(name) < 4:
        print_error_result("Cluster name should be at least 4 characters long")
        return

    ########################################
    # validate pre-creation conditions
    ########################################
    if _check_helm_availablity() == False:
        return
    print_info("Cluster installation dependencies are met.")

    config_path, context = _try_load_kubeconfig()
    if config_path is None:
        return
    print_info("Kubernetes configuration is loaded")
    print_data({"Kubeconfig path": config_path, "Current kubernetes context": context})

    if _vessl_already_exist(kubernetes_namespace):
        if not prompt_confirm(
            f"VESSL agent is already installed on current Kubernetes cluster. Do you want to remove and re-install?",
            default=False,
        ):
            return
        else:
            print_info("Cluster already has VESSL agent installed but force continuing...")
            print_warning("Removing existing VESSL agent...")
            remove_suceess = _try_remove_vessl_agent(kubernetes_namespace)
            if not remove_suceess:
                return

    else:
        print_info("VESSL cluster not found in your environment. Installing VESSL cluster...")

    ########################################
    # prepare helm repo
    ########################################
    if not prompt_confirm(
        f"Adding VESSL helm repo: {VESSL_HELM_REPO}. Continue?",
        default=True,
    ):
        print_error_result("Cluster creation aborted")
        print_info("You can add VESSL helm repo manually by running:")
        print_info(f"helm repo add vessl {VESSL_HELM_REPO}")
        return
    add_and_update_helm_repo()

    input_data = {
        "VESSL cluster name": name,
        "Provider": provider,
        "Kubeconfig path": config_path,
        "Current kubernetes context": context,
        "Kubernetes namespace": kubernetes_namespace,
    }

    ########################################
    # build create cluster param
    ########################################
    create_cluster_param = _build_create_cluster_param(
        name,
        kubernetes_namespace,
        provider,
        extra_helm_values,
        config_path,
        organization_name,
        input_data,
    )

    ########################################
    # user confirmation
    ########################################
    print_info("Cluster creation options:")
    print_data(input_data)
    if not prompt_confirm(
        f"Are you sure you want to install VESSL agent on current Kubernetes cluster with options displayed above?",
        default=True,
    ):
        print_error_result("Cluster creation aborted")
        return

    print_info("Creating cluster...")
    vessl_installation_info = create_cluster(create_cluster_param)
    VesslConfigLoader().kubernetes_namespace = kubernetes_namespace
    print_success_result(f"cluster {name} created!")
    print_cluster_data(vessl_installation_info)


def _build_create_cluster_param(
    name,
    kubernetes_namespace,
    provider,
    extra_helm_values,
    config_path,
    organization_name,
    input_data,
) -> CreateClusterParam:
    """Builds CreateClusterParam object from user inputs.\n
    Conditionally adds storage class, local path provisioner, and pod resource path.\n
    Also, sets input data for user confirmation.

    """
    create_cluster_param = CreateClusterParam()
    create_cluster_param.cluster_name = name
    create_cluster_param.kubernetes_namespace = kubernetes_namespace
    create_cluster_param.provider = provider
    create_cluster_param.kubeconfig_path = config_path
    if organization_name:
        create_cluster_param.organization_name = organization_name

    if provider == "gcp":
        print_debug(
            f"(provider=gcp) Setting default storage class for GCP as [standart-rwo]",
        )
        input_data["default storage class (gcp)"] = "standard-rwo"
        create_cluster_param.agent_local_storage_class_name = "standard-rwo"
    if provider == "aws":
        print_debug(
            f"(provider=aws) Setting default storage class for AWS as [ebs-sc]",
        )
        input_data["default storage class (aws)"] = "ebs-sc"
        create_cluster_param.agent_local_storage_class_name = "ebs-sc"

    should_enable_local_path_provisioner = True if (provider == "on-premise") else False
    print_debug(
        f"{'Enable' if should_enable_local_path_provisioner else 'Disable'} local-path-provisioner for {provider} cluster.\n"
        + "  This will allow VESSL agent to use host local storage for your workspace.\n"
    )
    input_data[f"local-path-provisioner"] = should_enable_local_path_provisioner
    if not should_enable_local_path_provisioner:
        # Default enabled for on-prem, disabled for cloud
        create_cluster_param.enable_local_path_provisioner = False

    if extra_helm_values and len(extra_helm_values) > 0:
        print_debug(
            f"Extra Helm values will be passed to VESSL agent installation: {extra_helm_values}",
        )
        input_data["Extra Helm values"] = ", ".join(extra_helm_values)
        create_cluster_param.extra_helm_values = extra_helm_values

    # dcgm patch
    v1 = k8sclient.CoreV1Api()
    # @XXX(seokju) rough assumption: k0s user will install k0s nodes
    # if this becomes a problem, patch chart and cli.
    node = v1.list_node().items[0]
    if "k0s" in node.status.node_info.kubelet_version:
        print_debug(
            f"(Required) Detected k0s node. Patching dcgm exporter podResourcePath as /var/lib/k0s/kubelete/pod-resources.",
        )
        input_data["(k0s) Patch dcgm exporter podResourcePath"] = True
        create_cluster_param.pod_resource_path = "/var/lib/k0s/kubelet/pod-resources"
    return create_cluster_param


def _vessl_already_exist(namespace: str):
    print_info("Checking existing VESSL installation...")
    try:
        vessl_installation_info = _check_existing_vessl_installation(namespace=namespace)
        if vessl_installation_info is not None:
            print_success("Existing VESSL installation found.")
            print_cluster_data(vessl_installation_info)
            return True
        return False
    except InvalidKernelClusterError as e:
        if e.code == ACCESS_TOKEN_NOT_FOUND_ERROR_CODE:
            print_error_result("Existing VESSL installation found, but access token is missing.")
        else:
            print_error_result(f"Existing VESSL installation found, but it is invalid. {repr(e)}")
        return True
    except VesslApiException as e:
        if e.status == 404:
            print_error(
                f"Existing VESSL installation, but not registered in VESSL Service: [{repr(e)}]"
            )
        else:
            print_error_result(
                f"Existing VESSL installation, but failed to check registration on VESSL Service: {repr(e)}"
            )
        return True


def _try_remove_vessl_agent(namespace: str) -> bool:
    config_path, context = _try_load_kubeconfig()
    if config_path is None:
        print_error_result("Failed to load kubeconfig")
        return False
    stream = os.popen(f"helm uninstall vessl --namespace {namespace} --kubeconfig {config_path}")
    output = stream.read()
    print_info(output)
    if stream.close() != None:
        print_error_result("Failed to uninstall vessl-agent")
        return False
    return True


def _try_load_kubeconfig():
    print_info("Loading cluster kubeconfig...")
    loaded, (config_path, context) = _load_cluster_kubeconfig()
    if not loaded:
        print_error_result("Kubernetes cluster not found;")
        print_error("Please configure valid kubeconfig file location by KUBECONFIG envvar.")
        print_info(
            "If you are a test user and want to install k0s-in-docker on your local machine, please run 'vessl cluster init-local' first."
        )
        return None, None

    return config_path, context


def _check_helm_availablity():
    print_debug("Checking if helm exists...")
    if which("helm") is None:
        print_error_result(
            "Command 'helm' not found; Please install Helm to install VESSL cluster agent: https://helm.sh/"
        )
        return False
    print_debug("helm command available!")
    return True


@cli.vessl_command()
@vessl_argument("name", type=click.STRING, required=True, prompter=cluster_name_prompter)
@organization_name_option
def read(name: str):
    cluster = read_cluster(cluster_name=name)
    print_cluster_data(cluster)


@cli.vessl_command()
@organization_name_option
def list():
    clusters = list_clusters()
    print_table(
        clusters,
        ["ID", "Name", "Type", "Status", "K8s Master Endpoint", "K8s Namespace"],
        lambda x: [
            x.id,
            x.name,
            "Managed" if x.is_savvihub_managed else "Custom",
            x.status.replace("-", " "),
            format_string(x.name),
            format_string(x.kubernetes_namespace),
        ],
    )


@cli.vessl_command()
@vessl_argument("name", type=click.STRING, required=True, prompter=custom_cluster_name_prompter)
@organization_name_option
def delete(name: str):
    if not prompt_confirm(f"Are you sure you want to delete '{name}'?"):
        print_error_result("Cluster deletion aborted")
        return
    cluster = read_cluster(cluster_name=name)
    data = delete_cluster(cluster_id=cluster.id)
    print_success(f"Deleted '{name}'.")

    config = VesslConfigLoader()

    if prompt_confirm(f"Do you want to destroy the cluster installed with VESSL CLI?"):
        namespace = config.kubernetes_namespace
        if namespace is None:
            print_error_result(
                "Failed to find kubernetes namespace. Please check your kubeconfig file."
            )
        else:
            if not _try_remove_vessl_agent(namespace=namespace):
                print_error(
                    "Failed to remove vessl-agent. Please remove it manually by 'helm uninstall vessl --namespace <namespace>'"
                )
            else:
                print_success("vessl-agent removed successfully.")

    if config.use_k0s_in_docker:
        if _check_k0s_exists():
            print_info(
                "It seems that you are using k0s-in-docker installed by `vessl cluster init-local`."
            )
            if prompt_confirm(
                "Do you want to remove k0s-in-docker? (note that this is NOT RECOVERABLE)"
            ):
                is_error = _try_remove_existing_k0s()
                if is_error:
                    print_error_result("Failed to remove k0s-in-docker.")
                    print_info(
                        f"You can try execute the following in the shell (note that this is NOT RECOVERABLE)."
                    )
                    print_info(f"$ docker stop vessl-k0s && docker rm vessl-k0s")
                    print_info(
                        f'$ docker run -it --rm --privileged --pid=host alpine nsenter -t 1 -m -- sh -c "rm -rfv /var/lib/k0s"'
                    )
        else:
            config.use_k0s_in_docker = False


@cli.vessl_command()
@vessl_argument("name", type=click.STRING, required=True, prompter=custom_cluster_name_prompter)
@vessl_argument("new_name", type=click.STRING, required=True, prompter=generic_prompter("New name"))
@organization_name_option
def rename(name: str, new_name: str):
    cluster = read_cluster(cluster_name=name)
    cluster = rename_cluster(cluster_id=cluster.id, new_cluster_name=new_name)
    print_success(f"Renamed '{name}' to '{cluster.name}'.")


@cli.vessl_command()
@vessl_argument("name", type=click.STRING, required=True, prompter=custom_cluster_name_prompter)
@organization_name_option
def list_nodes(name: str):
    cluster = read_cluster(cluster_name=name)
    cluster_nodes = list_cluster_nodes(cluster_id=cluster.id)
    print_table(
        cluster_nodes,
        ["ID", "Name", "CPU Limits", "GPU Limits", "Memory Limits"],
        lambda x: [x.id, x.name, x.cpu_limits, x.gpu_limits, x.memory_limits],
    )


@cli.vessl_command()
def init_local():
    print_info("Checking if docker exists...")
    try:
        docker.from_env()
    except docker.errors.DockerException as e:
        print_error(
            f"[{repr(e)}] Docker not found; Please install Docker to run VESSL on your machine.: https://docs.docker.com/engine/install/"
        )
        return
    print_info("docker command available!")

    print_info("Checking if k0s-in-docker already exists...")
    if _check_k0s_exists():
        if prompt_confirm("k0s-in-docker already exists. Do you want to remove and reinstall it?"):
            is_error = _try_remove_existing_k0s()
            if is_error:
                print_error_result("Failed to remove k0s-in-docker.")
                return
            print_success("k0s-in-docker removed successfully.")
        else:
            print_info("You need to remove existing local cluster first.")
            print_info("Aborted.")
            return

    if not prompt_confirm(
        "Are you sure you want to initialize local cluster with k0s on docker?",
    ):
        print_info("Aborted.")
        return

    print_info(
        "Creating local cluster with k0s... (More Info: https://docs.k0sproject.io/v1.25.12+k0s.0/k0s-in-docker)"
    )
    kubeconfig, context_name, already_exists = _install_k0s_in_docker()
    if kubeconfig is None:
        print_error("Failed to create local cluster!")
        return
    if already_exists:
        print_info("Cluster already exists")
    else:
        print_info("Local cluster created!")

    VesslConfigLoader().use_k0s_in_docker = True
    print_success(
        "Initializing cluster success!, Hit `vessl cluster create` to rocket your Machine Learning!",
    )


def _check_k0s_exists():
    client = docker.from_env()
    containers = client.containers.list(filters={"name": "vessl-k0s"}, all=True)
    if len(containers) > 0:
        print_info("k0s-in-docker already exists!")
        print_data(
            {"Id": containers[0].id, "Name": containers[0].name, "Status": containers[0].status}
        )
        return True
    print_info("k0s-in-docker does not exist!")
    return False


def _try_remove_existing_k0s():
    client = docker.from_env()
    containers = client.containers.list(filters={"name": "vessl-k0s"}, all=True)
    is_error = False
    print_info("Removing k0s-in-docker...")
    try:
        for container in containers:
            if container.status == "running":
                container.stop()
            container.remove(force=True)
    except docker.errors.APIError as e:
        print_error(f"Failed to uninstall k0s-in-docker!: {repr(e)}")
        is_error = True
        return is_error
    print_info("Removing connected volume...")
    try:
        stream = os.popen(
            'docker run -it --rm --privileged --pid=host alpine nsenter -t 1 -m -- sh -c "rm -rfv /var/lib/k0s"'
        )
        output = stream.read()
        if stream.close() != None:
            raise Exception(output)
        print_info(output)
    except Exception as e:
        print_error(f"Failed to remove connected volume!: {repr(e)}")
        print_info("You can try run following command manually;")
        print_info(
            '> docker run -it --rm --privileged --pid=host alpine nsenter -t 1 -m -- sh -c "rm -rfv /var/lib/k0s"'
        )
        is_error = True
        return is_error
    print_info("Uninstalled k0s-in-docker!")

    print_info("Removing custom kubeconfig...")
    try:
        kubeconfig_path = VesslConfigLoader().cluster_kubeconfig
        if os.path.exists(kubeconfig_path):
            os.remove(kubeconfig_path)
    except Exception as e:
        print_error(f"Failed to remove custom kubeconfig!: {repr(e)}")
        is_error = True
        return is_error
    print_info("Removed custom kubeconfig!")

    return is_error


def _install_k0s_in_docker() -> Tuple[str, str, bool]:
    """Install k0s-in-Docker on local machine.
    More Info: https://docs.k0sproject.io/v1.25.12+k0s.0/k0s-in-docker/
    """
    print_info(f"Installing k0s-in-docker...")
    k0s_config = """apiVersion: k0s.k0sproject.io/v1beta1
kind: ClusterConfig
metadata:
  name: k0s
spec:
  api:
    extraArgs:
      service-node-port-range: 32639-32767
"""
    kwargs = {
        "name": "vessl-k0s",
        "environment": {"K0S_CONFIG": k0s_config},
        "hostname": f"vessl-k0s-{socket.gethostname()}",
        "ports": {
            "6443/tcp": 6443,
        },
        "privileged": True,
        "volumes": {
            "/var/lib/k0s": {
                "bind": "/var/lib/k0s",
                "mode": "rw",
            },
        },
    }
    for i in range(32639, 32768):
        kwargs["ports"][f"{i}/tcp"] = i

    # Add extra arguments for arm64 machines
    if uname().machine == "arm64":
        print_info("System is using CPU with arm64 architecture; adding additional kwargs")
        kwargs["environment"]["ETCD_UNSUPPORTED_ARCH"] = "arm64"

    client = docker.from_env()

    # Add extra arguments for docker container running on Docker Desktop
    is_docker_desktop = client.info().get("OperatingSystem", "") == "Docker Desktop"
    if is_docker_desktop:
        print_info("System is running on Docker Desktop")

    if prompt_confirm(
        f"Should mount cgroup from host? (default: {'Y' if is_docker_desktop else 'N'}, Recommended for Docker Desktop)",
        default=is_docker_desktop,
    ):
        kwargs["cgroupns"] = "host"
        kwargs["volumes"]["/sys/fs/cgroup"] = {
            "bind": "/sys/fs/cgroup",
            "mode": "rw",
        }

    already_exists = False

    # Run k0s container
    print_info("k0s container configuration: ")
    print_warning(json.dumps(kwargs, indent=2))
    if not prompt_confirm("Starting k0s in docker with upper configuraion; Continue?"):
        print_info("Aborted.")
        return None, None, already_exists
    container = None
    try:
        container = _start_k0s_container(kwargs)
    except docker.errors.APIError as e:
        if e.status_code == 409 and "vessl-k0s" in e.explanation:
            print_error(
                "k0s-in-docker container already exists.",
            )
            already_exists = True
        else:
            print_error_result(
                f"[{repr(e)}] Failed to start k0s-in-docker container. Please check your docker configuration.",
            )
            return None, None, already_exists
    if already_exists:
        try:
            container = _get_running_k0s_container()
        except:
            print_error_result(
                "Failed to get running k0s-in-docker container. Please remove the existing container and try again.",
            )
            click.Abort()
            return None, None, already_exists

    # Wait for the container to be ready
    print_info(f"Waiting for the k0s-in-docker node to be ready...")
    try:
        _wait_for_k0s_ready(container)
    except Exception as e:
        click.secho(
            f"[{repr(e)}]; Error while waiting. Failed to initialize k0s-in-docker node.",
            bg="red",
            err=True,
        )
        click.Abort()
        return None, None, already_exists
    print_info(f"k0s-in-docker node is ready!")

    # Get kubeconfig, keep at /opt/vessl/cluster_kubeconfig.yaml for later
    print_info(f"Acquiring kubeconfig...")
    kubeconfig_filename = _acquire_kubeconfig(container)
    current_context = _load_kubeconfig_to_config(kubeconfig_filename)
    print_info(
        f"k0s-in-docker is ready! Saved kubeconfig at {kubeconfig_filename} for later use.",
    )
    return kubeconfig_filename, current_context["name"], already_exists


def cluster_id_prompter(ctx: click.Context, param: click.Parameter, value: str) -> int:
    clusters = list_clusters()
    cluster = prompt_choices(
        "Cluster",
        [(f"{x.name}", x) for x in clusters],
    )
    ctx.obj["cluster"] = cluster
    return cluster.id


def cluster_name_prompter(ctx: click.Context, param: click.Parameter, value: str) -> str:
    clusters = list_clusters()
    cluster = prompt_choices(
        "Cluster",
        [(f"{x.name}", x) for x in clusters],
    )
    ctx.obj["cluster"] = cluster
    return cluster.name


def cluster_name_callback(ctx: click.Context, param: click.Parameter, value: str) -> str:
    if "cluster" not in ctx.obj:
        ctx.obj["cluster"] = read_cluster(value)
    return value


def print_cluster_data(cluster: ResponseKernelClusterInfo, **kwargs):
    print_data(
        {
            "Name": cluster.name,
            "Dashboard URL": Endpoint.cluster.format(_get_organization_name(**kwargs), cluster.id),
            "Type": "VESSL-managed" if cluster.is_savvihub_managed else "Custom",
            "Region": format_string(cluster.region),
            "Status": cluster.status.replace("-", " "),
            "K8s Master Endpoint": format_string(cluster.name),
            "K8s Namespace": format_string(cluster.kubernetes_namespace),
            "K8s Service Type": cluster.kubernetes_service_type,
        }
    )


cluster_option = vessl_option(
    "-c",
    "--cluster",
    type=click.STRING,
    required=True,
    prompter=cluster_name_prompter,
    callback=cluster_name_callback,
    help="Must be specified before resource-related options (`--resource`, `--processor`, ...).",
)



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from openapi_client.configuration import Configuration


class ResponseWorkloadInfo(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'created_by': 'ResponseSimpleUser',
        'created_dt': 'datetime',
        'experiment': 'ResponseSimpleExperiment',
        'model_service_revision': 'ResponseModelServiceRevision',
        'project': 'ResponseSimpleProject',
        'run_execution': 'ResponseReducedRunExecution',
        'status': 'str',
        'type': 'str',
        'workspace': 'ResponseSimpleWorkspace'
    }

    attribute_map = {
        'created_by': 'created_by',
        'created_dt': 'created_dt',
        'experiment': 'experiment',
        'model_service_revision': 'model_service_revision',
        'project': 'project',
        'run_execution': 'run_execution',
        'status': 'status',
        'type': 'type',
        'workspace': 'workspace'
    }

    def __init__(self, created_by=None, created_dt=None, experiment=None, model_service_revision=None, project=None, run_execution=None, status=None, type=None, workspace=None, local_vars_configuration=None):  # noqa: E501
        """ResponseWorkloadInfo - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._created_by = None
        self._created_dt = None
        self._experiment = None
        self._model_service_revision = None
        self._project = None
        self._run_execution = None
        self._status = None
        self._type = None
        self._workspace = None
        self.discriminator = None

        self.created_by = created_by
        self.created_dt = created_dt
        if experiment is not None:
            self.experiment = experiment
        if model_service_revision is not None:
            self.model_service_revision = model_service_revision
        self.project = project
        if run_execution is not None:
            self.run_execution = run_execution
        if status is not None:
            self.status = status
        if type is not None:
            self.type = type
        if workspace is not None:
            self.workspace = workspace

    @property
    def created_by(self):
        """Gets the created_by of this ResponseWorkloadInfo.  # noqa: E501


        :return: The created_by of this ResponseWorkloadInfo.  # noqa: E501
        :rtype: ResponseSimpleUser
        """
        return self._created_by

    @created_by.setter
    def created_by(self, created_by):
        """Sets the created_by of this ResponseWorkloadInfo.


        :param created_by: The created_by of this ResponseWorkloadInfo.  # noqa: E501
        :type created_by: ResponseSimpleUser
        """
        if self.local_vars_configuration.client_side_validation and created_by is None:  # noqa: E501
            raise ValueError("Invalid value for `created_by`, must not be `None`")  # noqa: E501

        self._created_by = created_by

    @property
    def created_dt(self):
        """Gets the created_dt of this ResponseWorkloadInfo.  # noqa: E501


        :return: The created_dt of this ResponseWorkloadInfo.  # noqa: E501
        :rtype: datetime
        """
        return self._created_dt

    @created_dt.setter
    def created_dt(self, created_dt):
        """Sets the created_dt of this ResponseWorkloadInfo.


        :param created_dt: The created_dt of this ResponseWorkloadInfo.  # noqa: E501
        :type created_dt: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_dt is None:  # noqa: E501
            raise ValueError("Invalid value for `created_dt`, must not be `None`")  # noqa: E501

        self._created_dt = created_dt

    @property
    def experiment(self):
        """Gets the experiment of this ResponseWorkloadInfo.  # noqa: E501


        :return: The experiment of this ResponseWorkloadInfo.  # noqa: E501
        :rtype: ResponseSimpleExperiment
        """
        return self._experiment

    @experiment.setter
    def experiment(self, experiment):
        """Sets the experiment of this ResponseWorkloadInfo.


        :param experiment: The experiment of this ResponseWorkloadInfo.  # noqa: E501
        :type experiment: ResponseSimpleExperiment
        """

        self._experiment = experiment

    @property
    def model_service_revision(self):
        """Gets the model_service_revision of this ResponseWorkloadInfo.  # noqa: E501


        :return: The model_service_revision of this ResponseWorkloadInfo.  # noqa: E501
        :rtype: ResponseModelServiceRevision
        """
        return self._model_service_revision

    @model_service_revision.setter
    def model_service_revision(self, model_service_revision):
        """Sets the model_service_revision of this ResponseWorkloadInfo.


        :param model_service_revision: The model_service_revision of this ResponseWorkloadInfo.  # noqa: E501
        :type model_service_revision: ResponseModelServiceRevision
        """

        self._model_service_revision = model_service_revision

    @property
    def project(self):
        """Gets the project of this ResponseWorkloadInfo.  # noqa: E501


        :return: The project of this ResponseWorkloadInfo.  # noqa: E501
        :rtype: ResponseSimpleProject
        """
        return self._project

    @project.setter
    def project(self, project):
        """Sets the project of this ResponseWorkloadInfo.


        :param project: The project of this ResponseWorkloadInfo.  # noqa: E501
        :type project: ResponseSimpleProject
        """
        if self.local_vars_configuration.client_side_validation and project is None:  # noqa: E501
            raise ValueError("Invalid value for `project`, must not be `None`")  # noqa: E501

        self._project = project

    @property
    def run_execution(self):
        """Gets the run_execution of this ResponseWorkloadInfo.  # noqa: E501


        :return: The run_execution of this ResponseWorkloadInfo.  # noqa: E501
        :rtype: ResponseReducedRunExecution
        """
        return self._run_execution

    @run_execution.setter
    def run_execution(self, run_execution):
        """Sets the run_execution of this ResponseWorkloadInfo.


        :param run_execution: The run_execution of this ResponseWorkloadInfo.  # noqa: E501
        :type run_execution: ResponseReducedRunExecution
        """

        self._run_execution = run_execution

    @property
    def status(self):
        """Gets the status of this ResponseWorkloadInfo.  # noqa: E501


        :return: The status of this ResponseWorkloadInfo.  # noqa: E501
        :rtype: str
        """
        return self._status

    @status.setter
    def status(self, status):
        """Sets the status of this ResponseWorkloadInfo.


        :param status: The status of this ResponseWorkloadInfo.  # noqa: E501
        :type status: str
        """

        self._status = status

    @property
    def type(self):
        """Gets the type of this ResponseWorkloadInfo.  # noqa: E501


        :return: The type of this ResponseWorkloadInfo.  # noqa: E501
        :rtype: str
        """
        return self._type

    @type.setter
    def type(self, type):
        """Sets the type of this ResponseWorkloadInfo.


        :param type: The type of this ResponseWorkloadInfo.  # noqa: E501
        :type type: str
        """

        self._type = type

    @property
    def workspace(self):
        """Gets the workspace of this ResponseWorkloadInfo.  # noqa: E501


        :return: The workspace of this ResponseWorkloadInfo.  # noqa: E501
        :rtype: ResponseSimpleWorkspace
        """
        return self._workspace

    @workspace.setter
    def workspace(self, workspace):
        """Sets the workspace of this ResponseWorkloadInfo.


        :param workspace: The workspace of this ResponseWorkloadInfo.  # noqa: E501
        :type workspace: ResponseSimpleWorkspace
        """

        self._workspace = workspace

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ResponseWorkloadInfo):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ResponseWorkloadInfo):
            return True

        return self.to_dict() != other.to_dict()



import os
from typing import Dict, List, Optional, SupportsFloat, Union

from openapi_client.models import ResponseExperimentInfo
from vessl import vessl_api
from vessl.experiment import read_experiment_by_id
from vessl.util.image import Image

ImageRowType = Dict[str, List[Image]]
MetricRowType = Dict[str, SupportsFloat]
RowType = Union[ImageRowType, MetricRowType]

current_experiment: Optional[ResponseExperimentInfo] = None


def _update_current_experiment():
    global current_experiment
    experiment_id = os.environ.get("VESSL_EXPERIMENT_ID", None)
    access_token = os.environ.get("VESSL_ACCESS_TOKEN", None)

    if experiment_id is None or access_token is None:
        return

    vessl_api.configure_access_token(access_token)
    current_experiment = read_experiment_by_id(int(experiment_id))


def get_current_experiment() -> Optional[ResponseExperimentInfo]:
    global current_experiment
    if current_experiment != None:
        return current_experiment

    _update_current_experiment()
    return current_experiment


def get_current_experiment_id() -> int:
    assert current_experiment
    assert current_experiment.id
    return current_experiment.id



import os
import pathlib
import tempfile
from typing import Dict, Generator, List, Tuple

import pytest

from vessl.util.vesslignore import list_files_using_vesslignore


@pytest.fixture()
def vesslignore_test_files(tmp_path: pathlib.Path):
    """
    Pytest fixture function that prepares and removes files for .vesslignore test.
    """

    ignore_contents: Dict[str, List[str]] = {}
    files_to_write: List[Tuple[str, bool]] = []  # (path, should_exist)

    def _add_ignore(pattern, write_to=".vesslignore"):
        if write_to not in ignore_contents:
            ignore_contents[write_to] = []
        ignore_contents[write_to].append(pattern)

    def _expect_present(filename):
        nonlocal files_to_write
        files_to_write.append((filename, True))

    def _expect_ignored(filename):
        nonlocal files_to_write
        files_to_write.append((filename, False))

    # Comment
    _add_ignore("# Comments should work")

    # Simple bans
    _add_ignore("*.o")
    _add_ignore(".*.sw?")  # Vim swap files -- to test '?'
    _expect_present("SHOWN")
    _expect_ignored("HIDDEN.o")
    _expect_ignored(".HIDDEN.swp")
    _expect_ignored(".HIDDEN.swo")
    _expect_present("SHOWN.obj")  # do not allow the partial match 'SHOWN.o'
    # Simple bans in subdirectory
    _expect_present("data1/SHOWN")
    _expect_ignored("data1/HIDDEN.o")
    _expect_ignored("data1/.HIDDEN.swp")
    _expect_ignored("data1/.HIDDEN.swo")
    # Simple bans should match directory names, too
    _add_ignore("HIDDEN_DIR?")
    _expect_ignored("data1/HIDDEN_DIR3/HIDDEN.txt")

    # Patterns with '**/' should match in all places
    _add_ignore("**/animal/cat")
    _expect_present("data2/animal/dog")
    _expect_ignored("data2/animal/cat")
    _expect_ignored("data2/dist/animal/cat/mango")  # directory match
    _expect_present("data2/animal/catto")  # basename should match exactly, not prefixed
    # Trailing slash should force directory match
    _add_ignore("**/fruit/apples/")
    _expect_present("data2/fruit/apples")
    _expect_ignored("data2/dist/fruit/apples/Macintosh")

    # Relative path from root
    _add_ignore("data3/instrument/viol*")
    _expect_present("data3/instrument/trumpet")
    _expect_ignored("data3/instrument/violin")
    _expect_ignored("data3/instrument/viola/3")  # directory match
    # wildcard inside path
    _add_ignore("data3/f?sh/*/tail")
    _expect_present("data3/fish/tuna/head")
    _expect_ignored("data3/fish/tuna/tail")
    _expect_present("data3/fish/tuna/tailor")  # basename should not be prefixed
    # Trailing slash should force directory match
    _add_ignore("data3/fish/*/head/")
    _expect_present("data3/fish/cod/heading")
    _expect_present("data3/fish/cod/head")
    _expect_ignored("data3/fish/octopus/head/eye")

    # .vesslignore in subdirectory
    def _add_ignore_sub(pattern):
        _add_ignore(pattern, "subdir/.vesslignore")

    # Subdirectory pattern test
    _add_ignore_sub("cat")
    _add_ignore_sub("**/SHOWN")
    _add_ignore_sub("SUB_IGNORE/")
    _expect_ignored("subdir/cat")
    _expect_present("subdir/catto")
    _expect_ignored("subdir/data1/HIDDEN.o")
    _expect_present("subdir/data2/SUB_IGNORE/blob")
    _expect_present("subdir/data3/SUB_IGNORE")
    _expect_ignored("subdir/SUB_IGNORE/blob")
    _expect_present("SUB_IGNORE/SHOWN")  # subdir pattern should not affect parent

    for filename, present in files_to_write:
        path = tmp_path / filename
        path.parent.mkdir(exist_ok=True, parents=True)
        path.touch()
        pass

    for ignore_filename, ignore_lines in ignore_contents.items():
        path = tmp_path / ignore_filename
        path.parent.mkdir(exist_ok=True, parents=True)

        ignore_content = "\n".join(ignore_lines) + "\n"
        path.open("w").write(ignore_content)

    yield (tmp_path, files_to_write)


class TestVesslIgnore:
    def test_list_files_using_vesslignore(
        self, vesslignore_test_files: Tuple[pathlib.Path, List[Tuple[str, bool]]]
    ):
        root_path, files_to_write = vesslignore_test_files

        files_found = list_files_using_vesslignore(str(root_path))

        for filename, should_exist in files_to_write:
            assert (filename in files_found) == should_exist



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from openapi_client.configuration import Configuration


class ResponsePipeline(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'created_dt': 'datetime',
        'description': 'str',
        'last_triggered_dt': 'datetime',
        'last_triggered_reason': 'str',
        'name': 'str',
        'recent_executions': 'list[ResponseReducedPipelineExecution]',
        'spec': 'ResponsePipelineSpec',
        'triggers': 'list[ResponsePipelineTrigger]',
        'updated_dt': 'datetime'
    }

    attribute_map = {
        'created_dt': 'created_dt',
        'description': 'description',
        'last_triggered_dt': 'last_triggered_dt',
        'last_triggered_reason': 'last_triggered_reason',
        'name': 'name',
        'recent_executions': 'recent_executions',
        'spec': 'spec',
        'triggers': 'triggers',
        'updated_dt': 'updated_dt'
    }

    def __init__(self, created_dt=None, description=None, last_triggered_dt=None, last_triggered_reason=None, name=None, recent_executions=None, spec=None, triggers=None, updated_dt=None, local_vars_configuration=None):  # noqa: E501
        """ResponsePipeline - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._created_dt = None
        self._description = None
        self._last_triggered_dt = None
        self._last_triggered_reason = None
        self._name = None
        self._recent_executions = None
        self._spec = None
        self._triggers = None
        self._updated_dt = None
        self.discriminator = None

        self.created_dt = created_dt
        self.description = description
        self.last_triggered_dt = last_triggered_dt
        self.last_triggered_reason = last_triggered_reason
        self.name = name
        self.recent_executions = recent_executions
        if spec is not None:
            self.spec = spec
        self.triggers = triggers
        self.updated_dt = updated_dt

    @property
    def created_dt(self):
        """Gets the created_dt of this ResponsePipeline.  # noqa: E501


        :return: The created_dt of this ResponsePipeline.  # noqa: E501
        :rtype: datetime
        """
        return self._created_dt

    @created_dt.setter
    def created_dt(self, created_dt):
        """Sets the created_dt of this ResponsePipeline.


        :param created_dt: The created_dt of this ResponsePipeline.  # noqa: E501
        :type created_dt: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_dt is None:  # noqa: E501
            raise ValueError("Invalid value for `created_dt`, must not be `None`")  # noqa: E501

        self._created_dt = created_dt

    @property
    def description(self):
        """Gets the description of this ResponsePipeline.  # noqa: E501


        :return: The description of this ResponsePipeline.  # noqa: E501
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """Sets the description of this ResponsePipeline.


        :param description: The description of this ResponsePipeline.  # noqa: E501
        :type description: str
        """
        if self.local_vars_configuration.client_side_validation and description is None:  # noqa: E501
            raise ValueError("Invalid value for `description`, must not be `None`")  # noqa: E501

        self._description = description

    @property
    def last_triggered_dt(self):
        """Gets the last_triggered_dt of this ResponsePipeline.  # noqa: E501


        :return: The last_triggered_dt of this ResponsePipeline.  # noqa: E501
        :rtype: datetime
        """
        return self._last_triggered_dt

    @last_triggered_dt.setter
    def last_triggered_dt(self, last_triggered_dt):
        """Sets the last_triggered_dt of this ResponsePipeline.


        :param last_triggered_dt: The last_triggered_dt of this ResponsePipeline.  # noqa: E501
        :type last_triggered_dt: datetime
        """

        self._last_triggered_dt = last_triggered_dt

    @property
    def last_triggered_reason(self):
        """Gets the last_triggered_reason of this ResponsePipeline.  # noqa: E501


        :return: The last_triggered_reason of this ResponsePipeline.  # noqa: E501
        :rtype: str
        """
        return self._last_triggered_reason

    @last_triggered_reason.setter
    def last_triggered_reason(self, last_triggered_reason):
        """Sets the last_triggered_reason of this ResponsePipeline.


        :param last_triggered_reason: The last_triggered_reason of this ResponsePipeline.  # noqa: E501
        :type last_triggered_reason: str
        """

        self._last_triggered_reason = last_triggered_reason

    @property
    def name(self):
        """Gets the name of this ResponsePipeline.  # noqa: E501


        :return: The name of this ResponsePipeline.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this ResponsePipeline.


        :param name: The name of this ResponsePipeline.  # noqa: E501
        :type name: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def recent_executions(self):
        """Gets the recent_executions of this ResponsePipeline.  # noqa: E501


        :return: The recent_executions of this ResponsePipeline.  # noqa: E501
        :rtype: list[ResponseReducedPipelineExecution]
        """
        return self._recent_executions

    @recent_executions.setter
    def recent_executions(self, recent_executions):
        """Sets the recent_executions of this ResponsePipeline.


        :param recent_executions: The recent_executions of this ResponsePipeline.  # noqa: E501
        :type recent_executions: list[ResponseReducedPipelineExecution]
        """
        if self.local_vars_configuration.client_side_validation and recent_executions is None:  # noqa: E501
            raise ValueError("Invalid value for `recent_executions`, must not be `None`")  # noqa: E501

        self._recent_executions = recent_executions

    @property
    def spec(self):
        """Gets the spec of this ResponsePipeline.  # noqa: E501


        :return: The spec of this ResponsePipeline.  # noqa: E501
        :rtype: ResponsePipelineSpec
        """
        return self._spec

    @spec.setter
    def spec(self, spec):
        """Sets the spec of this ResponsePipeline.


        :param spec: The spec of this ResponsePipeline.  # noqa: E501
        :type spec: ResponsePipelineSpec
        """

        self._spec = spec

    @property
    def triggers(self):
        """Gets the triggers of this ResponsePipeline.  # noqa: E501


        :return: The triggers of this ResponsePipeline.  # noqa: E501
        :rtype: list[ResponsePipelineTrigger]
        """
        return self._triggers

    @triggers.setter
    def triggers(self, triggers):
        """Sets the triggers of this ResponsePipeline.


        :param triggers: The triggers of this ResponsePipeline.  # noqa: E501
        :type triggers: list[ResponsePipelineTrigger]
        """
        if self.local_vars_configuration.client_side_validation and triggers is None:  # noqa: E501
            raise ValueError("Invalid value for `triggers`, must not be `None`")  # noqa: E501

        self._triggers = triggers

    @property
    def updated_dt(self):
        """Gets the updated_dt of this ResponsePipeline.  # noqa: E501


        :return: The updated_dt of this ResponsePipeline.  # noqa: E501
        :rtype: datetime
        """
        return self._updated_dt

    @updated_dt.setter
    def updated_dt(self, updated_dt):
        """Sets the updated_dt of this ResponsePipeline.


        :param updated_dt: The updated_dt of this ResponsePipeline.  # noqa: E501
        :type updated_dt: datetime
        """
        if self.local_vars_configuration.client_side_validation and updated_dt is None:  # noqa: E501
            raise ValueError("Invalid value for `updated_dt`, must not be `None`")  # noqa: E501

        self._updated_dt = updated_dt

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ResponsePipeline):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ResponsePipeline):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from openapi_client.configuration import Configuration


class ResponseDatasetSource(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'bucket_name': 'str',
        'is_public_bucket': 'bool',
        'is_vessl_managed': 'bool',
        'on_premise_volume_config': 'OrmOnPremiseVolumeConfig',
        'path': 'str',
        'provider': 'str'
    }

    attribute_map = {
        'bucket_name': 'bucket_name',
        'is_public_bucket': 'is_public_bucket',
        'is_vessl_managed': 'is_vessl_managed',
        'on_premise_volume_config': 'on_premise_volume_config',
        'path': 'path',
        'provider': 'provider'
    }

    def __init__(self, bucket_name=None, is_public_bucket=None, is_vessl_managed=None, on_premise_volume_config=None, path=None, provider=None, local_vars_configuration=None):  # noqa: E501
        """ResponseDatasetSource - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._bucket_name = None
        self._is_public_bucket = None
        self._is_vessl_managed = None
        self._on_premise_volume_config = None
        self._path = None
        self._provider = None
        self.discriminator = None

        self.bucket_name = bucket_name
        self.is_public_bucket = is_public_bucket
        self.is_vessl_managed = is_vessl_managed
        if on_premise_volume_config is not None:
            self.on_premise_volume_config = on_premise_volume_config
        self.path = path
        self.provider = provider

    @property
    def bucket_name(self):
        """Gets the bucket_name of this ResponseDatasetSource.  # noqa: E501


        :return: The bucket_name of this ResponseDatasetSource.  # noqa: E501
        :rtype: str
        """
        return self._bucket_name

    @bucket_name.setter
    def bucket_name(self, bucket_name):
        """Sets the bucket_name of this ResponseDatasetSource.


        :param bucket_name: The bucket_name of this ResponseDatasetSource.  # noqa: E501
        :type bucket_name: str
        """

        self._bucket_name = bucket_name

    @property
    def is_public_bucket(self):
        """Gets the is_public_bucket of this ResponseDatasetSource.  # noqa: E501


        :return: The is_public_bucket of this ResponseDatasetSource.  # noqa: E501
        :rtype: bool
        """
        return self._is_public_bucket

    @is_public_bucket.setter
    def is_public_bucket(self, is_public_bucket):
        """Sets the is_public_bucket of this ResponseDatasetSource.


        :param is_public_bucket: The is_public_bucket of this ResponseDatasetSource.  # noqa: E501
        :type is_public_bucket: bool
        """
        if self.local_vars_configuration.client_side_validation and is_public_bucket is None:  # noqa: E501
            raise ValueError("Invalid value for `is_public_bucket`, must not be `None`")  # noqa: E501

        self._is_public_bucket = is_public_bucket

    @property
    def is_vessl_managed(self):
        """Gets the is_vessl_managed of this ResponseDatasetSource.  # noqa: E501


        :return: The is_vessl_managed of this ResponseDatasetSource.  # noqa: E501
        :rtype: bool
        """
        return self._is_vessl_managed

    @is_vessl_managed.setter
    def is_vessl_managed(self, is_vessl_managed):
        """Sets the is_vessl_managed of this ResponseDatasetSource.


        :param is_vessl_managed: The is_vessl_managed of this ResponseDatasetSource.  # noqa: E501
        :type is_vessl_managed: bool
        """
        if self.local_vars_configuration.client_side_validation and is_vessl_managed is None:  # noqa: E501
            raise ValueError("Invalid value for `is_vessl_managed`, must not be `None`")  # noqa: E501

        self._is_vessl_managed = is_vessl_managed

    @property
    def on_premise_volume_config(self):
        """Gets the on_premise_volume_config of this ResponseDatasetSource.  # noqa: E501


        :return: The on_premise_volume_config of this ResponseDatasetSource.  # noqa: E501
        :rtype: OrmOnPremiseVolumeConfig
        """
        return self._on_premise_volume_config

    @on_premise_volume_config.setter
    def on_premise_volume_config(self, on_premise_volume_config):
        """Sets the on_premise_volume_config of this ResponseDatasetSource.


        :param on_premise_volume_config: The on_premise_volume_config of this ResponseDatasetSource.  # noqa: E501
        :type on_premise_volume_config: OrmOnPremiseVolumeConfig
        """

        self._on_premise_volume_config = on_premise_volume_config

    @property
    def path(self):
        """Gets the path of this ResponseDatasetSource.  # noqa: E501


        :return: The path of this ResponseDatasetSource.  # noqa: E501
        :rtype: str
        """
        return self._path

    @path.setter
    def path(self, path):
        """Sets the path of this ResponseDatasetSource.


        :param path: The path of this ResponseDatasetSource.  # noqa: E501
        :type path: str
        """

        self._path = path

    @property
    def provider(self):
        """Gets the provider of this ResponseDatasetSource.  # noqa: E501


        :return: The provider of this ResponseDatasetSource.  # noqa: E501
        :rtype: str
        """
        return self._provider

    @provider.setter
    def provider(self, provider):
        """Sets the provider of this ResponseDatasetSource.


        :param provider: The provider of this ResponseDatasetSource.  # noqa: E501
        :type provider: str
        """
        if self.local_vars_configuration.client_side_validation and provider is None:  # noqa: E501
            raise ValueError("Invalid value for `provider`, must not be `None`")  # noqa: E501

        self._provider = provider

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ResponseDatasetSource):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ResponseDatasetSource):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from openapi_client.configuration import Configuration


class ResponseModelServiceInfo(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'alive_revision_numbers': 'list[int]',
        'created_by': 'ResponseUser',
        'created_dt': 'datetime',
        'description': 'str',
        'gateway_config': 'ResponseModelServiceGatewayInfo',
        'id': 'int',
        'kernel_cluster': 'ResponseKernelCluster',
        'kernel_cluster_id': 'int',
        'kernel_cluster_name': 'str',
        'last_updated_by': 'ResponseUser',
        'name': 'str',
        'running_replicas': 'int',
        'status': 'str',
        'total_replicas': 'int',
        'updated_dt': 'datetime'
    }

    attribute_map = {
        'alive_revision_numbers': 'alive_revision_numbers',
        'created_by': 'created_by',
        'created_dt': 'created_dt',
        'description': 'description',
        'gateway_config': 'gateway_config',
        'id': 'id',
        'kernel_cluster': 'kernel_cluster',
        'kernel_cluster_id': 'kernel_cluster_id',
        'kernel_cluster_name': 'kernel_cluster_name',
        'last_updated_by': 'last_updated_by',
        'name': 'name',
        'running_replicas': 'running_replicas',
        'status': 'status',
        'total_replicas': 'total_replicas',
        'updated_dt': 'updated_dt'
    }

    def __init__(self, alive_revision_numbers=None, created_by=None, created_dt=None, description=None, gateway_config=None, id=None, kernel_cluster=None, kernel_cluster_id=None, kernel_cluster_name=None, last_updated_by=None, name=None, running_replicas=None, status=None, total_replicas=None, updated_dt=None, local_vars_configuration=None):  # noqa: E501
        """ResponseModelServiceInfo - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._alive_revision_numbers = None
        self._created_by = None
        self._created_dt = None
        self._description = None
        self._gateway_config = None
        self._id = None
        self._kernel_cluster = None
        self._kernel_cluster_id = None
        self._kernel_cluster_name = None
        self._last_updated_by = None
        self._name = None
        self._running_replicas = None
        self._status = None
        self._total_replicas = None
        self._updated_dt = None
        self.discriminator = None

        self.alive_revision_numbers = alive_revision_numbers
        self.created_by = created_by
        self.created_dt = created_dt
        self.description = description
        if gateway_config is not None:
            self.gateway_config = gateway_config
        self.id = id
        self.kernel_cluster = kernel_cluster
        self.kernel_cluster_id = kernel_cluster_id
        self.kernel_cluster_name = kernel_cluster_name
        self.last_updated_by = last_updated_by
        self.name = name
        self.running_replicas = running_replicas
        self.status = status
        self.total_replicas = total_replicas
        self.updated_dt = updated_dt

    @property
    def alive_revision_numbers(self):
        """Gets the alive_revision_numbers of this ResponseModelServiceInfo.  # noqa: E501


        :return: The alive_revision_numbers of this ResponseModelServiceInfo.  # noqa: E501
        :rtype: list[int]
        """
        return self._alive_revision_numbers

    @alive_revision_numbers.setter
    def alive_revision_numbers(self, alive_revision_numbers):
        """Sets the alive_revision_numbers of this ResponseModelServiceInfo.


        :param alive_revision_numbers: The alive_revision_numbers of this ResponseModelServiceInfo.  # noqa: E501
        :type alive_revision_numbers: list[int]
        """
        if self.local_vars_configuration.client_side_validation and alive_revision_numbers is None:  # noqa: E501
            raise ValueError("Invalid value for `alive_revision_numbers`, must not be `None`")  # noqa: E501

        self._alive_revision_numbers = alive_revision_numbers

    @property
    def created_by(self):
        """Gets the created_by of this ResponseModelServiceInfo.  # noqa: E501


        :return: The created_by of this ResponseModelServiceInfo.  # noqa: E501
        :rtype: ResponseUser
        """
        return self._created_by

    @created_by.setter
    def created_by(self, created_by):
        """Sets the created_by of this ResponseModelServiceInfo.


        :param created_by: The created_by of this ResponseModelServiceInfo.  # noqa: E501
        :type created_by: ResponseUser
        """
        if self.local_vars_configuration.client_side_validation and created_by is None:  # noqa: E501
            raise ValueError("Invalid value for `created_by`, must not be `None`")  # noqa: E501

        self._created_by = created_by

    @property
    def created_dt(self):
        """Gets the created_dt of this ResponseModelServiceInfo.  # noqa: E501


        :return: The created_dt of this ResponseModelServiceInfo.  # noqa: E501
        :rtype: datetime
        """
        return self._created_dt

    @created_dt.setter
    def created_dt(self, created_dt):
        """Sets the created_dt of this ResponseModelServiceInfo.


        :param created_dt: The created_dt of this ResponseModelServiceInfo.  # noqa: E501
        :type created_dt: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_dt is None:  # noqa: E501
            raise ValueError("Invalid value for `created_dt`, must not be `None`")  # noqa: E501

        self._created_dt = created_dt

    @property
    def description(self):
        """Gets the description of this ResponseModelServiceInfo.  # noqa: E501


        :return: The description of this ResponseModelServiceInfo.  # noqa: E501
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """Sets the description of this ResponseModelServiceInfo.


        :param description: The description of this ResponseModelServiceInfo.  # noqa: E501
        :type description: str
        """
        if self.local_vars_configuration.client_side_validation and description is None:  # noqa: E501
            raise ValueError("Invalid value for `description`, must not be `None`")  # noqa: E501

        self._description = description

    @property
    def gateway_config(self):
        """Gets the gateway_config of this ResponseModelServiceInfo.  # noqa: E501


        :return: The gateway_config of this ResponseModelServiceInfo.  # noqa: E501
        :rtype: ResponseModelServiceGatewayInfo
        """
        return self._gateway_config

    @gateway_config.setter
    def gateway_config(self, gateway_config):
        """Sets the gateway_config of this ResponseModelServiceInfo.


        :param gateway_config: The gateway_config of this ResponseModelServiceInfo.  # noqa: E501
        :type gateway_config: ResponseModelServiceGatewayInfo
        """

        self._gateway_config = gateway_config

    @property
    def id(self):
        """Gets the id of this ResponseModelServiceInfo.  # noqa: E501


        :return: The id of this ResponseModelServiceInfo.  # noqa: E501
        :rtype: int
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this ResponseModelServiceInfo.


        :param id: The id of this ResponseModelServiceInfo.  # noqa: E501
        :type id: int
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def kernel_cluster(self):
        """Gets the kernel_cluster of this ResponseModelServiceInfo.  # noqa: E501


        :return: The kernel_cluster of this ResponseModelServiceInfo.  # noqa: E501
        :rtype: ResponseKernelCluster
        """
        return self._kernel_cluster

    @kernel_cluster.setter
    def kernel_cluster(self, kernel_cluster):
        """Sets the kernel_cluster of this ResponseModelServiceInfo.


        :param kernel_cluster: The kernel_cluster of this ResponseModelServiceInfo.  # noqa: E501
        :type kernel_cluster: ResponseKernelCluster
        """
        if self.local_vars_configuration.client_side_validation and kernel_cluster is None:  # noqa: E501
            raise ValueError("Invalid value for `kernel_cluster`, must not be `None`")  # noqa: E501

        self._kernel_cluster = kernel_cluster

    @property
    def kernel_cluster_id(self):
        """Gets the kernel_cluster_id of this ResponseModelServiceInfo.  # noqa: E501


        :return: The kernel_cluster_id of this ResponseModelServiceInfo.  # noqa: E501
        :rtype: int
        """
        return self._kernel_cluster_id

    @kernel_cluster_id.setter
    def kernel_cluster_id(self, kernel_cluster_id):
        """Sets the kernel_cluster_id of this ResponseModelServiceInfo.


        :param kernel_cluster_id: The kernel_cluster_id of this ResponseModelServiceInfo.  # noqa: E501
        :type kernel_cluster_id: int
        """
        if self.local_vars_configuration.client_side_validation and kernel_cluster_id is None:  # noqa: E501
            raise ValueError("Invalid value for `kernel_cluster_id`, must not be `None`")  # noqa: E501

        self._kernel_cluster_id = kernel_cluster_id

    @property
    def kernel_cluster_name(self):
        """Gets the kernel_cluster_name of this ResponseModelServiceInfo.  # noqa: E501


        :return: The kernel_cluster_name of this ResponseModelServiceInfo.  # noqa: E501
        :rtype: str
        """
        return self._kernel_cluster_name

    @kernel_cluster_name.setter
    def kernel_cluster_name(self, kernel_cluster_name):
        """Sets the kernel_cluster_name of this ResponseModelServiceInfo.


        :param kernel_cluster_name: The kernel_cluster_name of this ResponseModelServiceInfo.  # noqa: E501
        :type kernel_cluster_name: str
        """
        if self.local_vars_configuration.client_side_validation and kernel_cluster_name is None:  # noqa: E501
            raise ValueError("Invalid value for `kernel_cluster_name`, must not be `None`")  # noqa: E501

        self._kernel_cluster_name = kernel_cluster_name

    @property
    def last_updated_by(self):
        """Gets the last_updated_by of this ResponseModelServiceInfo.  # noqa: E501


        :return: The last_updated_by of this ResponseModelServiceInfo.  # noqa: E501
        :rtype: ResponseUser
        """
        return self._last_updated_by

    @last_updated_by.setter
    def last_updated_by(self, last_updated_by):
        """Sets the last_updated_by of this ResponseModelServiceInfo.


        :param last_updated_by: The last_updated_by of this ResponseModelServiceInfo.  # noqa: E501
        :type last_updated_by: ResponseUser
        """
        if self.local_vars_configuration.client_side_validation and last_updated_by is None:  # noqa: E501
            raise ValueError("Invalid value for `last_updated_by`, must not be `None`")  # noqa: E501

        self._last_updated_by = last_updated_by

    @property
    def name(self):
        """Gets the name of this ResponseModelServiceInfo.  # noqa: E501


        :return: The name of this ResponseModelServiceInfo.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this ResponseModelServiceInfo.


        :param name: The name of this ResponseModelServiceInfo.  # noqa: E501
        :type name: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def running_replicas(self):
        """Gets the running_replicas of this ResponseModelServiceInfo.  # noqa: E501


        :return: The running_replicas of this ResponseModelServiceInfo.  # noqa: E501
        :rtype: int
        """
        return self._running_replicas

    @running_replicas.setter
    def running_replicas(self, running_replicas):
        """Sets the running_replicas of this ResponseModelServiceInfo.


        :param running_replicas: The running_replicas of this ResponseModelServiceInfo.  # noqa: E501
        :type running_replicas: int
        """
        if self.local_vars_configuration.client_side_validation and running_replicas is None:  # noqa: E501
            raise ValueError("Invalid value for `running_replicas`, must not be `None`")  # noqa: E501

        self._running_replicas = running_replicas

    @property
    def status(self):
        """Gets the status of this ResponseModelServiceInfo.  # noqa: E501


        :return: The status of this ResponseModelServiceInfo.  # noqa: E501
        :rtype: str
        """
        return self._status

    @status.setter
    def status(self, status):
        """Sets the status of this ResponseModelServiceInfo.


        :param status: The status of this ResponseModelServiceInfo.  # noqa: E501
        :type status: str
        """
        if self.local_vars_configuration.client_side_validation and status is None:  # noqa: E501
            raise ValueError("Invalid value for `status`, must not be `None`")  # noqa: E501

        self._status = status

    @property
    def total_replicas(self):
        """Gets the total_replicas of this ResponseModelServiceInfo.  # noqa: E501


        :return: The total_replicas of this ResponseModelServiceInfo.  # noqa: E501
        :rtype: int
        """
        return self._total_replicas

    @total_replicas.setter
    def total_replicas(self, total_replicas):
        """Sets the total_replicas of this ResponseModelServiceInfo.


        :param total_replicas: The total_replicas of this ResponseModelServiceInfo.  # noqa: E501
        :type total_replicas: int
        """
        if self.local_vars_configuration.client_side_validation and total_replicas is None:  # noqa: E501
            raise ValueError("Invalid value for `total_replicas`, must not be `None`")  # noqa: E501

        self._total_replicas = total_replicas

    @property
    def updated_dt(self):
        """Gets the updated_dt of this ResponseModelServiceInfo.  # noqa: E501


        :return: The updated_dt of this ResponseModelServiceInfo.  # noqa: E501
        :rtype: datetime
        """
        return self._updated_dt

    @updated_dt.setter
    def updated_dt(self, updated_dt):
        """Sets the updated_dt of this ResponseModelServiceInfo.


        :param updated_dt: The updated_dt of this ResponseModelServiceInfo.  # noqa: E501
        :type updated_dt: datetime
        """
        if self.local_vars_configuration.client_side_validation and updated_dt is None:  # noqa: E501
            raise ValueError("Invalid value for `updated_dt`, must not be `None`")  # noqa: E501

        self._updated_dt = updated_dt

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ResponseModelServiceInfo):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ResponseModelServiceInfo):
            return True

        return self.to_dict() != other.to_dict()



from test.conftest import USE_MOCK

import pytest
from mock import Mock, patch

import vessl
from vessl.util.exception import InvalidWorkspaceError


@pytest.mark.skipif(USE_MOCK, reason="Does not run if mocking is used.")
class TestWorkspace:
    def test_read_workspace(self):
        # Must create first
        pass

    def test_list_workspaces(self):
        vessl.list_workspaces()


def test_backup_workspace():
    # Test fail
    config_instance_mock = Mock()
    config_instance_mock.workspace = None
    config_mock = Mock(return_value=config_instance_mock)

    with patch.object(vessl.workspace, "VesslConfigLoader", config_mock):
        with pytest.raises(InvalidWorkspaceError):
            vessl.backup_workspace()

    # TODO: Add test success


def test_restore_workspace():
    # Test fail
    config_instance_mock = Mock()
    config_instance_mock.workspace = None
    config_mock = Mock(return_value=config_instance_mock)

    with patch.object(vessl.workspace, "VesslConfigLoader", config_mock):
        with pytest.raises(InvalidWorkspaceError):
            vessl.restore_workspace()

    # TODO: Add test success


# def test_connect_workspace_ssh():
#     workspace_id = 1234
#     private_key_path = "/path/to/key"
#
#     # Test fail
#     workspace_mock = Mock()
#     workspace_mock.status = "terminated"
#
#     read_workspace_ = patch.object(
#         vessl.workspace, "read_workspace", return_value=workspace_mock
#     )
#
#     with read_workspace_:
#         with pytest.raises(InvalidWorkspaceError):
#             vessl.connect_workspace_ssh(private_key_path)
#
#     # Test success
#     port = 1234
#     hostname = "tcp.hostname.com"
#
#     workspace_mock = Mock()
#     workspace_mock.status = "running"
#     workspace_mock.endpoints.ssh.endpoint = f"tcp://{hostname}:{port}"
#
#     read_workspace_ = patch.object(
#         vessl.workspace, "read_workspace", return_value=workspace_mock
#     )
#     system_ = patch("vessl.workspace.os.system")
#
#     with read_workspace_, system_:
#         vessl.connect_workspace_ssh(private_key_path)
#         vessl.workspace.os.system.assert_called_with(
#             f"ssh -p {port} -i {private_key_path} vessl@{hostname}"
#         )


# def test_update_vscode_remote_ssh():
#     # Setup
#     ssh_config_path = "test/ssh_config"
#     with open(ssh_config_path, "w+"):
#         pass
#
#     private_key_path = "/path/to/key"
#     name = "vessl_workspace"
#     created_dt = datetime.datetime.now()
#     port = 1234
#     hostname = "tcp.hostname.com"
#     host = f"{name}-{int(created_dt.timestamp())}"
#
#     workspace_mock = Mock()
#     workspace_mock.name = name
#     workspace_mock.created_dt = created_dt
#     workspace_mock.endpoints.ssh.endpoint = f"tcp://{hostname}:{port}"
#
#     list_workspaces_ = patch.object(
#         vessl.workspace, "list_workspaces", return_value=[workspace_mock]
#     )
#     paramiko_ = patch.object(vessl.workspace, "paramiko")
#     ssh_config_format_ = patch.object(
#         vessl.workspace, "SSH_CONFIG_PATH", ssh_config_path
#     )
#
#     with list_workspaces_, paramiko_, ssh_config_format_:
#         vessl.update_vscode_remote_ssh(private_key_path)
#
#     with open(ssh_config_path, "r") as f:
#         assert f.read() == SSH_CONFIG_FORMAT.format(
#             host=host,
#             hostname=hostname,
#             port=port,
#         ) + f'    IdentityFile {private_key_path}\n'
#
#     # Teardown
#     try:
#         os.remove(ssh_config_path)
#     except OSError:
#         pass



"""
Serving command definition and delegation.
"""
import sys
from typing import List, Optional, TextIO

import click
import yaml

from openapi_client.models import ResponseModelServiceInfo
from openapi_client.models.response_kernel_cluster import ResponseKernelCluster
from vessl import vessl_api
from vessl.cli._base import VesslCommand, VesslGroup
from vessl.cli.serve.util import (
    list_http_ports,
    print_gateway,
    print_revision,
    validate_port_choice,
)
from vessl.serving import (
    create_revision_from_yaml,
    list_revisions,
    list_servings,
    read_gateway,
    read_revision,
    update_gateway_for_revision,
    update_gateway_from_yaml,
)
from vessl.util.exception import VesslApiException

from .command_options import (
    enable_gateway_if_off_option,
    serving_name_option,
    update_gateway_option,
    update_gateway_port_option,
    update_gateway_weight_option,
)

cli = VesslGroup("serve")


@cli.command("list", cls=VesslCommand)
def serving_list():
    """
    List servings in current organization.
    """
    servings: List[ResponseModelServiceInfo] = list_servings(
        organization=vessl_api.organization.name
    ).results

    print(f"{len(servings)} serving(s) found.\n")
    for serving in servings:
        kernel_cluster: ResponseKernelCluster = serving.kernel_cluster
        status: str = serving.status  # "ready", "running", "error"

        print(f"{serving.name} (cluster {kernel_cluster.name}): status {status.capitalize()}")


@cli.group("revision")
def cli_revision():
    """
    Root command for revision-related commands.
    """
    pass


@cli_revision.command("create", cls=VesslCommand)
@serving_name_option
@click.option(
    "-f",
    "--file",
    type=click.File("r"),
    required=True,
    help="Path to YAML file for serving revision definition.",
)
@update_gateway_option
@enable_gateway_if_off_option
@update_gateway_weight_option
@update_gateway_port_option
def revision_create_with_yaml(
    serving: ResponseModelServiceInfo,
    file: TextIO,
    update_gateway: bool,
    enable_gateway_if_off: bool,
    update_gateway_weight: Optional[int] = None,
    update_gateway_port: Optional[int] = None,
):
    """
    Create serving revision with spec written in YAML.
    """
    if not update_gateway and (
        update_gateway_weight is not None or update_gateway_port is not None
    ):
        print("Cannot specify traffic weight or port when not updating gateway.")
        sys.exit(1)

    yaml_body = file.read()
    yaml_obj = yaml.safe_load(yaml_body)

    if update_gateway:
        # Do as much validation as possible before actually creating revision.
        # weight check
        if update_gateway_weight is None:
            update_gateway_weight = 100
        elif not 1 <= update_gateway_weight <= 100:
            print(f"Invalid weight: {update_gateway_weight}% (must be between 1 and 100)")
            sys.exit(1)

        # port check
        if update_gateway_port is None:
            http_ports = list_http_ports(yaml_obj)
            if len(http_ports) != 1:
                print(
                    "Error: port for gateway was not specified, and could not automatically "
                    "determine which port to use.\n"
                    f"{len(http_ports)} port(s) was found: " + ", ".join(map(str, http_ports))
                )
                sys.exit(1)
            update_gateway_port = http_ports[0]
            print(f"Automatically choosing port {update_gateway_port} for gateway.")
        elif not 1 <= update_gateway_port <= 65535:
            print(f"Invalid port: {update_gateway_port}")
            sys.exit(1)
        else:
            validate_port_choice(yaml_obj, update_gateway_port)

        # gateway status check
        if not enable_gateway_if_off:
            gateway_current = read_gateway(
                organization=vessl_api.organization.name, serving_name=serving.name
            )
            if not gateway_current.enabled:
                print("Cannot update gateway because it is not enabled. Please enable it first.")
                print("NOTE (current status of gateway):")
                print_gateway(gateway_current)
                sys.exit(1)

    revision = create_revision_from_yaml(
        organization=vessl_api.organization.name, serving_name=serving.name, yaml_body=yaml_body
    )
    print(f"Successfully created revision in serving {serving.name}.\n")
    print_revision(revision)

    if update_gateway:
        gateway_updated = update_gateway_for_revision(
            vessl_api.organization.name,
            serving_name=serving.name,
            revision_number=revision.number,
            port=update_gateway_port,
            weight=update_gateway_weight,
        )
        print(f"Successfully updated gateway for revision #{revision.number}.\n")
        print_gateway(gateway_updated)
    else:
        print(
            "NOTE: Since --update-gateway option was not given, "
            "you cannot currently access this revision via gateway.\n\n"
            "Either use --update-gateway when creating revision, or update gateway manually."
        )


@cli_revision.command("show", cls=VesslCommand)
@serving_name_option
@click.option("--number", "-n", required=True, type=int, help="Number of revision.")
def revision_show(serving: ResponseModelServiceInfo, number: int):
    """
    Show current status and information about a serving revision.
    """
    try:
        revision = read_revision(
            organization=vessl_api.organization.name,
            serving_name=serving.name,
            revision_number=number,
        )
    except VesslApiException as e:
        print(f"Failed to read revision #{number} of serving {serving.name}: {e.message}")
        sys.exit(1)

    print_revision(revision, verbose=True)


@cli_revision.command("list", cls=VesslCommand)
@serving_name_option
def revision_list(serving: ResponseModelServiceInfo):
    """
    List all revisions.
    """
    try:
        revisions = list_revisions(
            organization=vessl_api.organization.name,
            serving_name=serving.name,
        )
    except VesslApiException as e:
        print(f"Failed to list revisions of serving {serving.name}: {e.message}")
        sys.exit(1)

    print(f"{len(revisions)} revision(s) found.\n")

    for i, revision in enumerate(revisions):
        if i > 0:
            print()
        print_revision(revision)


@cli.group("gateway")
def cli_gateway():
    """
    Root command for gateway-related commands.
    """
    pass


@cli_gateway.command("show", cls=VesslCommand)
@serving_name_option
def gateway_show(serving: ResponseModelServiceInfo):
    """
    Show current status of the gateway of a serving.
    """
    try:
        gateway = read_gateway(
            organization=vessl_api.organization.name,
            serving_name=serving.name,
        )
    except VesslApiException as e:
        print(f"Failed to read gateway of serving {serving.name}: {e.message}")
        sys.exit(1)

    print_gateway(gateway)


@cli_gateway.command("update", cls=VesslCommand)
@serving_name_option
@click.option(
    "-f",
    "--file",
    type=click.File("r"),
    required=True,
    help="Path to YAML file for serving revision definition.",
)
def gateway_update_with_yaml(serving: ResponseModelServiceInfo, file: TextIO):
    """
    Update serving gateway with spec written in YAML.
    """
    yaml_body = file.read()

    gateway = update_gateway_from_yaml(
        organization=vessl_api.organization.name, serving_name=serving.name, yaml_body=yaml_body
    )
    print(f"Successfully update gateway of serving {serving.name}.\n")
    print_gateway(gateway)



import base64
import os
import socket
from logging import Logger
from platform import uname
from time import sleep
from typing import List, Optional, Tuple

import click
import docker
from kubernetes import client as k8sclient
from kubernetes import config
from kubernetes.config import KUBE_CONFIG_DEFAULT_LOCATION

from openapi_client.models import (
    ClusterUpdateAPIInput,
    ResponseKernelClusterInfo,
    ResponseKernelClusterNodeInfo,
)
from vessl import vessl_api
from vessl.kernel_cluster_install_command_builder import ClusterInstallCommandBuilder
from vessl.organization import _get_organization_name
from vessl.util import logger
from vessl.util.config import DEFAULT_VESSL_DIR
from vessl.util.constant import VESSL_HELM_REPO
from vessl.util.exception import (
    ACCESS_TOKEN_NOT_FOUND_ERROR_CODE,
    ClusterNotFoundError,
    InvalidKernelClusterError,
    VesslApiException,
    VesslRuntimeException,
)


class CreateClusterParam:
    """Create cluster parameter for `vessl.create_cluster`.

    Attributes:
    - Required
        cluster_name(str): Cluster name.
        kubernetes_namespace(str): Kubernetes namespace to install VESSL agent.
            defaults to "vessl".
        extra_helm_values(list[str]): Helm values to pass to cluster install
            command. See https://github.com/vessl-ai/cluster-resources/blob/main/helm-chart/values.yaml
            for available Helm values.
        kubeconfig_path(str): Path to kubeconfig file.
        provider(str): Cluster provider. defaults to "on-premise"

    - Optional
        organization_name(str): Organization name. If not given, then use
            default organization.
        agent_local_storage_name(str): Name of local storage class for agent.
        pod_resource_path(str): Path to pod resource file.
        enable_local_path_provisioner(bool): Whether to enable local path provisioner.
    """

    def __init__(
        self,
        cluster_name: str = "",
        kubeconfig_path: str = "",
        extra_helm_values: List[str] = [],
        provider: str = "on-premise",
        kubernetes_namespace: str = "vessl",
        organization_name: Optional[str] = None,
        agent_local_storage_name: str = "",
        pod_resource_path: str = "",
        enable_local_path_provisioner: bool = True,
    ) -> None:
        self.cluster_name = cluster_name
        self.kubernetes_namespace = kubernetes_namespace
        self.extra_helm_values = extra_helm_values
        self.kubeconfig_path = kubeconfig_path
        self.provider = provider

        # optional
        self.organization_name: str = organization_name
        self.agent_local_storage_name = agent_local_storage_name
        self.pod_resource_path = pod_resource_path
        self.enable_local_path_provisioner = enable_local_path_provisioner


def create_cluster(
    param: CreateClusterParam,
) -> ResponseKernelClusterInfo:
    """Create a VESSL cluster by installing VESSL agent to given Kubernetes
    namespace. If you want to override the default organization, then pass
    `organization_name` to `param`.

    Args:
        param(CreateClusterParam): Create cluster parameter.

    Example:
        ```python
        vessl.install_cluster(
            param=vessl.CreateClusterParam(
                cluster_name="foo",
                ...
            ),
        )
        ```
    """
    agent_access_token = vessl_api.custom_cluster_key_api(
        organization_name=_get_organization_name(**{"organization_name": param.organization_name}),
    ).access_token

    logger.debug(f"cluster name: '{param.cluster_name}'")
    logger.debug(f"access token: '{agent_access_token}'")

    cluster_install_command = (
        ClusterInstallCommandBuilder()
        .with_kubeconfig(param.kubeconfig_path)
        .with_cluster_name(param.cluster_name)
        .with_provider_type(param.provider)
        .with_namespace(param.kubernetes_namespace)
        .with_access_token(agent_access_token)
        .with_local_path_provisioner_enabled(param.enable_local_path_provisioner)
        .with_helm_values(param.extra_helm_values)
        .with_pod_resource_path(param.pod_resource_path)
        .build()
    )

    stream = os.popen(cluster_install_command)
    output = stream.read()
    if stream.close() is not None:
        raise VesslRuntimeException(f"Failed to install VESSL agent: {output}")
    print(click.style(output, fg="blue"))

    logger.warn(
        click.style(
            "VESSL cluster agent installed. Waiting for the agent to be connected with VESSL...",
            fg="green",
        )
    )
    for attempts in range(18):
        sleep(10)
        try:
            return vessl_api.custom_cluster_connectivity_api(
                organization_name=_get_organization_name(
                    **{"organization_name": param.organization_name}
                ),
                agent_access_token=agent_access_token,
            )
        except VesslApiException as e:
            if e.code == "NotFound":
                continue
            raise e

    raise ClusterNotFoundError(
        "Timeout for checking agent connection; Please check agent log for more information."
    )


def add_and_update_helm_repo(logger: Logger = None):
    add_stream = os.popen(
        " ".join(
            [
                "helm",
                "repo",
                "add",
                "vessl",
                VESSL_HELM_REPO,
            ]
        )
    )
    repo_add_output = add_stream.read()
    if add_stream.close() != None:
        if logger:
            logger.debug("Failed to add helm repo")
        else:
            print(click.style("Failed to add helm repo", fg="red"))
        raise VesslRuntimeException(f"Failed to add helm repo: {repo_add_output}")

    update_stream = os.popen(" ".join(["helm", "repo", "update"]))
    repo_update_output = update_stream.read()

    if update_stream.close() != None:
        if logger:
            logger.debug("Failed to update helm repo")
        else:
            print(click.style("Failed to update helm repo", fg="red"))
        raise VesslRuntimeException(f"Failed to update helm repo: {repo_update_output}")

    # if no logger (ex. called from cli), print output
    if logger:
        logger.debug(repo_add_output)
        logger.debug(repo_update_output)
    else:
        print(click.style(repo_add_output, fg="blue"))
        print(click.style(repo_update_output, fg="blue"))


def _load_cluster_kubeconfig() -> Tuple[bool, Tuple[Optional[str], Optional[str]]]:
    """Find and load kubeconfig for VESSL cluster installation.

    Returns:
        True, if valid kubernetes config has loaded
        False, if no valid kubernetes config has found
    """
    cluster_kubeconfig = _get_cluster_kubeconfig()
    config_files = [cluster_kubeconfig] if cluster_kubeconfig else []
    config_files.append("/var/lib/k0s/pki/admin.conf")  # for k0s node as a worker
    config_files.append(os.path.expanduser(KUBE_CONFIG_DEFAULT_LOCATION))

    for config_file in config_files:
        try:
            config.load_kube_config(config_file)
            _, current_context = config.list_kube_config_contexts(config_file)
            logger.info(
                f"Using kubeconfig from file: {config_file} / context: {current_context['name']}"
            )
            return True, (config_file, current_context["name"])
        except config.config_exception.ConfigException as e:
            logger.warn(f"Cluster install: load kubeconfig from {config_file} failed: {repr(e)}")

    return False, (None, None)


def _install_k0s_in_docker() -> Tuple[str, str]:
    """Install k0s-in-Docker on local machine."""
    logger.info(f"Installing k0s-in-docker ...")
    k0s_config = """apiVersion: k0s.k0sproject.io/v1beta1
kind: ClusterConfig
metadata:
  name: k0s
spec:
  api:
    extraArgs:
      service-node-port-range: 32639-32767
"""
    kwargs = {
        "name": "vessl-k0s",
        "environment": {"K0S_CONFIG": k0s_config},
        "hostname": f"vessl-k0s-{socket.gethostname()}",
        "ports": {
            "6443/tcp": 6443,
        },
        "privileged": True,
        "volumes": {
            "/var/lib/k0s": {
                "bind": "/var/lib/k0s",
                "mode": "rw",
            },
        },
    }
    for i in range(32639, 32768):
        kwargs["ports"][f"{i}/tcp"] = i

    # Add extra arguments for arm64 machines
    if uname().machine == "arm64":
        logger.debug("System is using CPU with arm64 architecture; adding additional kwargs")
        kwargs["environment"]["ETCD_UNSUPPORTED_ARCH"] = "arm64"

    client = docker.from_env()

    # Add extra arguments for docker container running on Docker Desktop
    if client.info().get("OperatingSystem", "") == "Docker Desktop":
        logger.debug("Docker is running on Docker Desktop; adding additional kwargs")
        kwargs["cgroupns"] = "host"
        kwargs["volumes"]["/sys/fs/cgroup"] = {
            "bind": "/sys/fs/cgroup",
            "mode": "rw",
        }

    # Run k0s container
    container = _start_k0s_container(kwargs)

    # Wait for the container to be ready
    _wait_for_k0s_ready(container)

    # Get kubeconfig, keep at /opt/vessl/cluster_kubeconfig.yaml for later
    logger.debug("Acquiring kubeconfig")
    kubeconfig_filename = _acquire_kubeconfig(container)

    current_context = _load_kubeconfig_to_config(kubeconfig_filename)
    logger.info(f"k0s-in-docker is ready! Saved kubeconfig at {kubeconfig_filename} for later use.")
    return kubeconfig_filename, current_context["name"]


def _load_kubeconfig_to_config(kubeconfig_filename):
    vessl_api.config_loader.cluster_kubeconfig = kubeconfig_filename
    config.load_kube_config(kubeconfig_filename)
    _, current_context = config.list_kube_config_contexts(kubeconfig_filename)
    return current_context


def _acquire_kubeconfig(container):
    res = container.exec_run("cat /var/lib/k0s/pki/admin.conf", stream=False, demux=True)
    output, _ = res.output
    kubeconfig = output.decode("utf-8")
    kubeconfig_filename = f"{DEFAULT_VESSL_DIR}/cluster_kubeconfig.yaml"
    with open(kubeconfig_filename, "w") as f:
        f.write(kubeconfig)
    return kubeconfig_filename


def _wait_for_k0s_ready(container):
    node_ready = False
    trial_limit = 10
    for trial in range(trial_limit):
        logger.info(f"Checking for the k0s-in-docker node to be ready... ({trial+1}/{trial_limit})")
        res = container.exec_run(
            "bash -c 'k0s kubectl get nodes --no-headers=true | grep -v NotReady | wc -l'",
            stream=False,
            demux=True,
        )
        output, err = res.output
        if output:
            logger.debug(output.decode("utf-8"))
            nodes = output.decode("utf-8").strip()
            if nodes == "1":
                node_ready = True
                break
        if err:
            logger.debug(err.decode("utf-8"))
        sleep(10)
    if not node_ready:
        container.kill()
        raise VesslRuntimeException("Timeout while waiting for k0s-in-docker node. Please retry.")


def _start_k0s_container(kwargs):
    client = docker.from_env()
    container = client.containers.run(
        detach=True,
        image=f"docker.io/k0sproject/k0s:{K0S_VERSION}",
        command="k0s controller --single --no-taints --config /etc/k0s/config.yaml",
        **kwargs,
    )

    return container


def _get_running_k0s_container() -> docker.models.containers.Container:
    """Get k0s-in-Docker container."""
    client = docker.from_env()
    containers = client.containers.list(filters={"name": "vessl-k0s", "status": "running"})
    if len(containers) == 0:
        raise VesslRuntimeException("k0s-in-Docker container not found.")
    return containers[0]


def _check_existing_vessl_installation(
    namespace: str = "vessl", **kwargs
) -> ResponseKernelClusterInfo:
    """Check if there is existing vessl installation in current Kubernetes
    cluster."""

    # Find secret with name=vessl-agent, annotation=vessl
    try:
        v1 = k8sclient.CoreV1Api()
        vessl_secret: k8sclient.V1Secret = v1.read_namespaced_secret(
            name="vessl-agent", namespace=namespace
        )
    except k8sclient.ApiException as e:
        if e.status != 404:
            raise VesslRuntimeException("Error while communicating with kubernetes.")
        return None

    # VESSL agent token is found - there is existing cluster in current kubernetes cluster
    logger.warn("Existing VESSL cluster installation found! getting cluster information...")
    vessl_access_token = vessl_secret.data.get("access-token")
    if vessl_access_token is None:
        raise InvalidKernelClusterError(
            "Kubernetes secret for VESSL found, but secret does not have access-token",
            ACCESS_TOKEN_NOT_FOUND_ERROR_CODE,
        )
    vessl_access_token = base64.b64decode(vessl_access_token)
    logger.debug(f"access token: {vessl_access_token}")
    return vessl_api.custom_cluster_connectivity_api(
        organization_name=_get_organization_name(**kwargs),
        agent_access_token=vessl_access_token,
    )


def read_cluster(cluster_name: str, **kwargs) -> ResponseKernelClusterInfo:
    """Read cluster in the default organization. If you want to override the
    default organization, then pass `organization_name` as `**kwargs`.

    Args:
        cluster_name(str): Cluster name.

    Example:
        ```python
        vessl.read_cluster(
            cluster_name="seoul-cluster",
        )
        ```
    """
    kernel_clusters = list_clusters(**kwargs)
    kernel_clusters = {x.name: x for x in kernel_clusters}

    if cluster_name not in kernel_clusters:
        raise InvalidKernelClusterError(f"Kernel cluster not found: {cluster_name}")
    return kernel_clusters[cluster_name]


def list_clusters(**kwargs) -> List[ResponseKernelClusterInfo]:
    """List clusters in the default organization. If you want to override the
    default organization, then pass `organization_name` as `**kwargs`.

    Example:
        ```python
        vessl.list_clusters()
        ```
    """
    return vessl_api.cluster_list_api(
        organization_name=_get_organization_name(**kwargs),
    ).clusters


def delete_cluster(cluster_id: int, **kwargs) -> object:
    """Delete custom cluster in the default organization. If you want to
    override the default organization, then pass `organization_name` as
    `**kwargs`.

    Args:
        cluster_id(int): Cluster ID.

    Example:
        ```python
        vessl.delete_cluster(
            cluster_id=1,
        )
        ```
    """
    return vessl_api.cluster_delete_api(
        cluster_id=cluster_id,
        organization_name=_get_organization_name(**kwargs),
    )


def rename_cluster(cluster_id: int, new_cluster_name: str, **kwargs) -> ResponseKernelClusterInfo:
    """Rename custom cluster in the default organization. If you want to
    override the default organization, then pass `organization_name` as
    `**kwargs`.

    Args:
        cluster_id(int): Cluster ID.
        new_cluster_name(str): Cluster name to change.

    Example:
        ```python
        vessl.rename_cluster(
            cluster_id=1,
            new_cluster_name="seoul-cluster-2",
        )
        ```
    """
    return vessl_api.cluster_update_api(
        cluster_id=cluster_id,
        organization_name=_get_organization_name(**kwargs),
        custom_cluster_update_api_input=ClusterUpdateAPIInput(
            name=new_cluster_name,
        ),
    )


def list_cluster_nodes(cluster_id: int, **kwargs) -> List[ResponseKernelClusterNodeInfo]:
    """List custom cluster nodes in the default organization. If you want to
    override the default organization, then pass `organization_name` as
    `**kwargs`.

    Args:
        cluster_id(int): Cluster ID.

    Example:
        ```python
        vessl.list_cluster_nodes(
            cluster_id=1,
        )
        ```
    """
    return vessl_api.custom_cluster_node_list_api(
        cluster_id=cluster_id,
        organization_name=_get_organization_name(**kwargs),
    ).nodes


def _get_cluster_kubeconfig(**kwargs) -> str:
    cluster_kubeconfig = kwargs.get("cluster_kubeconfig")
    if cluster_kubeconfig is not None:
        return cluster_kubeconfig
    if vessl_api.cluster_kubeconfig is not None:
        return vessl_api.cluster_kubeconfig

    return ""



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from openapi_client.configuration import Configuration


class OrmWorkloadPodInfo(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'model_service_id': 'int',
        'node_name': 'str',
        'pod_ip': 'str',
        'pod_name': 'str',
        'revision_number': 'int',
        'status': 'str',
        'updated_dt': 'datetime'
    }

    attribute_map = {
        'model_service_id': 'model_service_id',
        'node_name': 'node_name',
        'pod_ip': 'pod_ip',
        'pod_name': 'pod_name',
        'revision_number': 'revision_number',
        'status': 'status',
        'updated_dt': 'updated_dt'
    }

    def __init__(self, model_service_id=None, node_name=None, pod_ip=None, pod_name=None, revision_number=None, status=None, updated_dt=None, local_vars_configuration=None):  # noqa: E501
        """OrmWorkloadPodInfo - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._model_service_id = None
        self._node_name = None
        self._pod_ip = None
        self._pod_name = None
        self._revision_number = None
        self._status = None
        self._updated_dt = None
        self.discriminator = None

        self.model_service_id = model_service_id
        self.node_name = node_name
        self.pod_ip = pod_ip
        self.pod_name = pod_name
        self.revision_number = revision_number
        self.status = status
        self.updated_dt = updated_dt

    @property
    def model_service_id(self):
        """Gets the model_service_id of this OrmWorkloadPodInfo.  # noqa: E501


        :return: The model_service_id of this OrmWorkloadPodInfo.  # noqa: E501
        :rtype: int
        """
        return self._model_service_id

    @model_service_id.setter
    def model_service_id(self, model_service_id):
        """Sets the model_service_id of this OrmWorkloadPodInfo.


        :param model_service_id: The model_service_id of this OrmWorkloadPodInfo.  # noqa: E501
        :type model_service_id: int
        """
        if self.local_vars_configuration.client_side_validation and model_service_id is None:  # noqa: E501
            raise ValueError("Invalid value for `model_service_id`, must not be `None`")  # noqa: E501

        self._model_service_id = model_service_id

    @property
    def node_name(self):
        """Gets the node_name of this OrmWorkloadPodInfo.  # noqa: E501


        :return: The node_name of this OrmWorkloadPodInfo.  # noqa: E501
        :rtype: str
        """
        return self._node_name

    @node_name.setter
    def node_name(self, node_name):
        """Sets the node_name of this OrmWorkloadPodInfo.


        :param node_name: The node_name of this OrmWorkloadPodInfo.  # noqa: E501
        :type node_name: str
        """
        if self.local_vars_configuration.client_side_validation and node_name is None:  # noqa: E501
            raise ValueError("Invalid value for `node_name`, must not be `None`")  # noqa: E501

        self._node_name = node_name

    @property
    def pod_ip(self):
        """Gets the pod_ip of this OrmWorkloadPodInfo.  # noqa: E501


        :return: The pod_ip of this OrmWorkloadPodInfo.  # noqa: E501
        :rtype: str
        """
        return self._pod_ip

    @pod_ip.setter
    def pod_ip(self, pod_ip):
        """Sets the pod_ip of this OrmWorkloadPodInfo.


        :param pod_ip: The pod_ip of this OrmWorkloadPodInfo.  # noqa: E501
        :type pod_ip: str
        """
        if self.local_vars_configuration.client_side_validation and pod_ip is None:  # noqa: E501
            raise ValueError("Invalid value for `pod_ip`, must not be `None`")  # noqa: E501

        self._pod_ip = pod_ip

    @property
    def pod_name(self):
        """Gets the pod_name of this OrmWorkloadPodInfo.  # noqa: E501


        :return: The pod_name of this OrmWorkloadPodInfo.  # noqa: E501
        :rtype: str
        """
        return self._pod_name

    @pod_name.setter
    def pod_name(self, pod_name):
        """Sets the pod_name of this OrmWorkloadPodInfo.


        :param pod_name: The pod_name of this OrmWorkloadPodInfo.  # noqa: E501
        :type pod_name: str
        """
        if self.local_vars_configuration.client_side_validation and pod_name is None:  # noqa: E501
            raise ValueError("Invalid value for `pod_name`, must not be `None`")  # noqa: E501

        self._pod_name = pod_name

    @property
    def revision_number(self):
        """Gets the revision_number of this OrmWorkloadPodInfo.  # noqa: E501


        :return: The revision_number of this OrmWorkloadPodInfo.  # noqa: E501
        :rtype: int
        """
        return self._revision_number

    @revision_number.setter
    def revision_number(self, revision_number):
        """Sets the revision_number of this OrmWorkloadPodInfo.


        :param revision_number: The revision_number of this OrmWorkloadPodInfo.  # noqa: E501
        :type revision_number: int
        """
        if self.local_vars_configuration.client_side_validation and revision_number is None:  # noqa: E501
            raise ValueError("Invalid value for `revision_number`, must not be `None`")  # noqa: E501

        self._revision_number = revision_number

    @property
    def status(self):
        """Gets the status of this OrmWorkloadPodInfo.  # noqa: E501


        :return: The status of this OrmWorkloadPodInfo.  # noqa: E501
        :rtype: str
        """
        return self._status

    @status.setter
    def status(self, status):
        """Sets the status of this OrmWorkloadPodInfo.


        :param status: The status of this OrmWorkloadPodInfo.  # noqa: E501
        :type status: str
        """
        if self.local_vars_configuration.client_side_validation and status is None:  # noqa: E501
            raise ValueError("Invalid value for `status`, must not be `None`")  # noqa: E501

        self._status = status

    @property
    def updated_dt(self):
        """Gets the updated_dt of this OrmWorkloadPodInfo.  # noqa: E501


        :return: The updated_dt of this OrmWorkloadPodInfo.  # noqa: E501
        :rtype: datetime
        """
        return self._updated_dt

    @updated_dt.setter
    def updated_dt(self, updated_dt):
        """Sets the updated_dt of this OrmWorkloadPodInfo.


        :param updated_dt: The updated_dt of this OrmWorkloadPodInfo.  # noqa: E501
        :type updated_dt: datetime
        """
        if self.local_vars_configuration.client_side_validation and updated_dt is None:  # noqa: E501
            raise ValueError("Invalid value for `updated_dt`, must not be `None`")  # noqa: E501

        self._updated_dt = updated_dt

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, OrmWorkloadPodInfo):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, OrmWorkloadPodInfo):
            return True

        return self.to_dict() != other.to_dict()



from vessl.internal.vessl_run import VesslRun

_run = VesslRun()

init = _run.init
api = _run.api
finish = _run.finish
log = _run.log
upload = _run.upload
progress = _run.progress
hyperparameters = _run.hyperparameters



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from openapi_client.configuration import Configuration


class ResponseRunSpec(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'cluster': 'ResponseKernelCluster',
        'description': 'str',
        'id': 'int',
        'kernel_image': 'ResponseKernelImage',
        'resource_spec_legacy': 'V1KernelResourceSpecFormLegacy',
        'resource_spec_preset_id': 'int',
        'spec': 'V1RunSpec',
        'title': 'str'
    }

    attribute_map = {
        'cluster': 'cluster',
        'description': 'description',
        'id': 'id',
        'kernel_image': 'kernel_image',
        'resource_spec_legacy': 'resource_spec_legacy',
        'resource_spec_preset_id': 'resource_spec_preset_id',
        'spec': 'spec',
        'title': 'title'
    }

    def __init__(self, cluster=None, description=None, id=None, kernel_image=None, resource_spec_legacy=None, resource_spec_preset_id=None, spec=None, title=None, local_vars_configuration=None):  # noqa: E501
        """ResponseRunSpec - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._cluster = None
        self._description = None
        self._id = None
        self._kernel_image = None
        self._resource_spec_legacy = None
        self._resource_spec_preset_id = None
        self._spec = None
        self._title = None
        self.discriminator = None

        if cluster is not None:
            self.cluster = cluster
        if description is not None:
            self.description = description
        if id is not None:
            self.id = id
        if kernel_image is not None:
            self.kernel_image = kernel_image
        if resource_spec_legacy is not None:
            self.resource_spec_legacy = resource_spec_legacy
        self.resource_spec_preset_id = resource_spec_preset_id
        if spec is not None:
            self.spec = spec
        if title is not None:
            self.title = title

    @property
    def cluster(self):
        """Gets the cluster of this ResponseRunSpec.  # noqa: E501


        :return: The cluster of this ResponseRunSpec.  # noqa: E501
        :rtype: ResponseKernelCluster
        """
        return self._cluster

    @cluster.setter
    def cluster(self, cluster):
        """Sets the cluster of this ResponseRunSpec.


        :param cluster: The cluster of this ResponseRunSpec.  # noqa: E501
        :type cluster: ResponseKernelCluster
        """

        self._cluster = cluster

    @property
    def description(self):
        """Gets the description of this ResponseRunSpec.  # noqa: E501


        :return: The description of this ResponseRunSpec.  # noqa: E501
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """Sets the description of this ResponseRunSpec.


        :param description: The description of this ResponseRunSpec.  # noqa: E501
        :type description: str
        """

        self._description = description

    @property
    def id(self):
        """Gets the id of this ResponseRunSpec.  # noqa: E501


        :return: The id of this ResponseRunSpec.  # noqa: E501
        :rtype: int
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this ResponseRunSpec.


        :param id: The id of this ResponseRunSpec.  # noqa: E501
        :type id: int
        """

        self._id = id

    @property
    def kernel_image(self):
        """Gets the kernel_image of this ResponseRunSpec.  # noqa: E501


        :return: The kernel_image of this ResponseRunSpec.  # noqa: E501
        :rtype: ResponseKernelImage
        """
        return self._kernel_image

    @kernel_image.setter
    def kernel_image(self, kernel_image):
        """Sets the kernel_image of this ResponseRunSpec.


        :param kernel_image: The kernel_image of this ResponseRunSpec.  # noqa: E501
        :type kernel_image: ResponseKernelImage
        """

        self._kernel_image = kernel_image

    @property
    def resource_spec_legacy(self):
        """Gets the resource_spec_legacy of this ResponseRunSpec.  # noqa: E501


        :return: The resource_spec_legacy of this ResponseRunSpec.  # noqa: E501
        :rtype: V1KernelResourceSpecFormLegacy
        """
        return self._resource_spec_legacy

    @resource_spec_legacy.setter
    def resource_spec_legacy(self, resource_spec_legacy):
        """Sets the resource_spec_legacy of this ResponseRunSpec.


        :param resource_spec_legacy: The resource_spec_legacy of this ResponseRunSpec.  # noqa: E501
        :type resource_spec_legacy: V1KernelResourceSpecFormLegacy
        """

        self._resource_spec_legacy = resource_spec_legacy

    @property
    def resource_spec_preset_id(self):
        """Gets the resource_spec_preset_id of this ResponseRunSpec.  # noqa: E501


        :return: The resource_spec_preset_id of this ResponseRunSpec.  # noqa: E501
        :rtype: int
        """
        return self._resource_spec_preset_id

    @resource_spec_preset_id.setter
    def resource_spec_preset_id(self, resource_spec_preset_id):
        """Sets the resource_spec_preset_id of this ResponseRunSpec.


        :param resource_spec_preset_id: The resource_spec_preset_id of this ResponseRunSpec.  # noqa: E501
        :type resource_spec_preset_id: int
        """

        self._resource_spec_preset_id = resource_spec_preset_id

    @property
    def spec(self):
        """Gets the spec of this ResponseRunSpec.  # noqa: E501


        :return: The spec of this ResponseRunSpec.  # noqa: E501
        :rtype: V1RunSpec
        """
        return self._spec

    @spec.setter
    def spec(self, spec):
        """Sets the spec of this ResponseRunSpec.


        :param spec: The spec of this ResponseRunSpec.  # noqa: E501
        :type spec: V1RunSpec
        """

        self._spec = spec

    @property
    def title(self):
        """Gets the title of this ResponseRunSpec.  # noqa: E501


        :return: The title of this ResponseRunSpec.  # noqa: E501
        :rtype: str
        """
        return self._title

    @title.setter
    def title(self, title):
        """Sets the title of this ResponseRunSpec.


        :param title: The title of this ResponseRunSpec.  # noqa: E501
        :type title: str
        """

        self._title = title

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ResponseRunSpec):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ResponseRunSpec):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from openapi_client.configuration import Configuration


class OrmAccessControlPolicy(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'created_dt': 'datetime',
        'edges': 'OrmAccessControlPolicyEdges',
        'group_id': 'int',
        'id': 'int',
        'organization_id': 'int',
        'permission': 'str',
        'project_id': 'int',
        'root_organization_id': 'int',
        'subject_everyone': 'bool',
        'updated_dt': 'datetime',
        'user_id': 'int'
    }

    attribute_map = {
        'created_dt': 'created_dt',
        'edges': 'edges',
        'group_id': 'group_id',
        'id': 'id',
        'organization_id': 'organization_id',
        'permission': 'permission',
        'project_id': 'project_id',
        'root_organization_id': 'root_organization_id',
        'subject_everyone': 'subject_everyone',
        'updated_dt': 'updated_dt',
        'user_id': 'user_id'
    }

    def __init__(self, created_dt=None, edges=None, group_id=None, id=None, organization_id=None, permission=None, project_id=None, root_organization_id=None, subject_everyone=None, updated_dt=None, user_id=None, local_vars_configuration=None):  # noqa: E501
        """OrmAccessControlPolicy - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._created_dt = None
        self._edges = None
        self._group_id = None
        self._id = None
        self._organization_id = None
        self._permission = None
        self._project_id = None
        self._root_organization_id = None
        self._subject_everyone = None
        self._updated_dt = None
        self._user_id = None
        self.discriminator = None

        if created_dt is not None:
            self.created_dt = created_dt
        if edges is not None:
            self.edges = edges
        self.group_id = group_id
        if id is not None:
            self.id = id
        self.organization_id = organization_id
        if permission is not None:
            self.permission = permission
        self.project_id = project_id
        self.root_organization_id = root_organization_id
        if subject_everyone is not None:
            self.subject_everyone = subject_everyone
        if updated_dt is not None:
            self.updated_dt = updated_dt
        self.user_id = user_id

    @property
    def created_dt(self):
        """Gets the created_dt of this OrmAccessControlPolicy.  # noqa: E501


        :return: The created_dt of this OrmAccessControlPolicy.  # noqa: E501
        :rtype: datetime
        """
        return self._created_dt

    @created_dt.setter
    def created_dt(self, created_dt):
        """Sets the created_dt of this OrmAccessControlPolicy.


        :param created_dt: The created_dt of this OrmAccessControlPolicy.  # noqa: E501
        :type created_dt: datetime
        """

        self._created_dt = created_dt

    @property
    def edges(self):
        """Gets the edges of this OrmAccessControlPolicy.  # noqa: E501


        :return: The edges of this OrmAccessControlPolicy.  # noqa: E501
        :rtype: OrmAccessControlPolicyEdges
        """
        return self._edges

    @edges.setter
    def edges(self, edges):
        """Sets the edges of this OrmAccessControlPolicy.


        :param edges: The edges of this OrmAccessControlPolicy.  # noqa: E501
        :type edges: OrmAccessControlPolicyEdges
        """

        self._edges = edges

    @property
    def group_id(self):
        """Gets the group_id of this OrmAccessControlPolicy.  # noqa: E501


        :return: The group_id of this OrmAccessControlPolicy.  # noqa: E501
        :rtype: int
        """
        return self._group_id

    @group_id.setter
    def group_id(self, group_id):
        """Sets the group_id of this OrmAccessControlPolicy.


        :param group_id: The group_id of this OrmAccessControlPolicy.  # noqa: E501
        :type group_id: int
        """

        self._group_id = group_id

    @property
    def id(self):
        """Gets the id of this OrmAccessControlPolicy.  # noqa: E501


        :return: The id of this OrmAccessControlPolicy.  # noqa: E501
        :rtype: int
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this OrmAccessControlPolicy.


        :param id: The id of this OrmAccessControlPolicy.  # noqa: E501
        :type id: int
        """

        self._id = id

    @property
    def organization_id(self):
        """Gets the organization_id of this OrmAccessControlPolicy.  # noqa: E501


        :return: The organization_id of this OrmAccessControlPolicy.  # noqa: E501
        :rtype: int
        """
        return self._organization_id

    @organization_id.setter
    def organization_id(self, organization_id):
        """Sets the organization_id of this OrmAccessControlPolicy.


        :param organization_id: The organization_id of this OrmAccessControlPolicy.  # noqa: E501
        :type organization_id: int
        """

        self._organization_id = organization_id

    @property
    def permission(self):
        """Gets the permission of this OrmAccessControlPolicy.  # noqa: E501


        :return: The permission of this OrmAccessControlPolicy.  # noqa: E501
        :rtype: str
        """
        return self._permission

    @permission.setter
    def permission(self, permission):
        """Sets the permission of this OrmAccessControlPolicy.


        :param permission: The permission of this OrmAccessControlPolicy.  # noqa: E501
        :type permission: str
        """

        self._permission = permission

    @property
    def project_id(self):
        """Gets the project_id of this OrmAccessControlPolicy.  # noqa: E501


        :return: The project_id of this OrmAccessControlPolicy.  # noqa: E501
        :rtype: int
        """
        return self._project_id

    @project_id.setter
    def project_id(self, project_id):
        """Sets the project_id of this OrmAccessControlPolicy.


        :param project_id: The project_id of this OrmAccessControlPolicy.  # noqa: E501
        :type project_id: int
        """

        self._project_id = project_id

    @property
    def root_organization_id(self):
        """Gets the root_organization_id of this OrmAccessControlPolicy.  # noqa: E501


        :return: The root_organization_id of this OrmAccessControlPolicy.  # noqa: E501
        :rtype: int
        """
        return self._root_organization_id

    @root_organization_id.setter
    def root_organization_id(self, root_organization_id):
        """Sets the root_organization_id of this OrmAccessControlPolicy.


        :param root_organization_id: The root_organization_id of this OrmAccessControlPolicy.  # noqa: E501
        :type root_organization_id: int
        """

        self._root_organization_id = root_organization_id

    @property
    def subject_everyone(self):
        """Gets the subject_everyone of this OrmAccessControlPolicy.  # noqa: E501


        :return: The subject_everyone of this OrmAccessControlPolicy.  # noqa: E501
        :rtype: bool
        """
        return self._subject_everyone

    @subject_everyone.setter
    def subject_everyone(self, subject_everyone):
        """Sets the subject_everyone of this OrmAccessControlPolicy.


        :param subject_everyone: The subject_everyone of this OrmAccessControlPolicy.  # noqa: E501
        :type subject_everyone: bool
        """

        self._subject_everyone = subject_everyone

    @property
    def updated_dt(self):
        """Gets the updated_dt of this OrmAccessControlPolicy.  # noqa: E501


        :return: The updated_dt of this OrmAccessControlPolicy.  # noqa: E501
        :rtype: datetime
        """
        return self._updated_dt

    @updated_dt.setter
    def updated_dt(self, updated_dt):
        """Sets the updated_dt of this OrmAccessControlPolicy.


        :param updated_dt: The updated_dt of this OrmAccessControlPolicy.  # noqa: E501
        :type updated_dt: datetime
        """

        self._updated_dt = updated_dt

    @property
    def user_id(self):
        """Gets the user_id of this OrmAccessControlPolicy.  # noqa: E501


        :return: The user_id of this OrmAccessControlPolicy.  # noqa: E501
        :rtype: int
        """
        return self._user_id

    @user_id.setter
    def user_id(self, user_id):
        """Sets the user_id of this OrmAccessControlPolicy.


        :param user_id: The user_id of this OrmAccessControlPolicy.  # noqa: E501
        :type user_id: int
        """

        self._user_id = user_id

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, OrmAccessControlPolicy):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, OrmAccessControlPolicy):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from openapi_client.configuration import Configuration


class OrmOrganizationEdges(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'billing_histories': 'list[OrmBillingHistory]',
        'cluster_quotas': 'list[OrmClusterQuota]',
        'credit_earn_histories': 'list[OrmCreditEarnHistory]',
        'default_storage': 'OrmStorage',
        'default_volume': 'OrmVolume',
        'groups': 'list[OrmGroup]',
        'kernel_clusters': 'list[OrmKernelCluster]',
        'model_services': 'list[OrmModelService]',
        'organization_credentials': 'list[OrmOrganizationCredentials]',
        'organization_kernel_clusters': 'list[OrmOrganizationKernelCluster]',
        'organization_policies': 'list[OrmAccessControlPolicy]',
        'owner': 'OrmUser',
        'pipeline_step_types': 'list[OrmPipelineStepType]',
        'pricing_plan': 'OrmPricingPlan',
        'primary_owner': 'OrmUser',
        'root_organization_policies': 'list[OrmAccessControlPolicy]',
        'stripe_billing_history': 'list[OrmStripeBillingHistory]',
        'user_organization': 'list[OrmUserOrganization]',
        'withdraw_history_organizations': 'list[OrmWithdrawHistoryOrganization]',
        'workloads': 'list[OrmWorkload]',
        'workspaces': 'list[OrmWorkspace]'
    }

    attribute_map = {
        'billing_histories': 'billing_histories',
        'cluster_quotas': 'cluster_quotas',
        'credit_earn_histories': 'credit_earn_histories',
        'default_storage': 'default_storage',
        'default_volume': 'default_volume',
        'groups': 'groups',
        'kernel_clusters': 'kernel_clusters',
        'model_services': 'model_services',
        'organization_credentials': 'organization_credentials',
        'organization_kernel_clusters': 'organization_kernel_clusters',
        'organization_policies': 'organization_policies',
        'owner': 'owner',
        'pipeline_step_types': 'pipeline_step_types',
        'pricing_plan': 'pricing_plan',
        'primary_owner': 'primary_owner',
        'root_organization_policies': 'root_organization_policies',
        'stripe_billing_history': 'stripe_billing_history',
        'user_organization': 'user_organization',
        'withdraw_history_organizations': 'withdraw_history_organizations',
        'workloads': 'workloads',
        'workspaces': 'workspaces'
    }

    def __init__(self, billing_histories=None, cluster_quotas=None, credit_earn_histories=None, default_storage=None, default_volume=None, groups=None, kernel_clusters=None, model_services=None, organization_credentials=None, organization_kernel_clusters=None, organization_policies=None, owner=None, pipeline_step_types=None, pricing_plan=None, primary_owner=None, root_organization_policies=None, stripe_billing_history=None, user_organization=None, withdraw_history_organizations=None, workloads=None, workspaces=None, local_vars_configuration=None):  # noqa: E501
        """OrmOrganizationEdges - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._billing_histories = None
        self._cluster_quotas = None
        self._credit_earn_histories = None
        self._default_storage = None
        self._default_volume = None
        self._groups = None
        self._kernel_clusters = None
        self._model_services = None
        self._organization_credentials = None
        self._organization_kernel_clusters = None
        self._organization_policies = None
        self._owner = None
        self._pipeline_step_types = None
        self._pricing_plan = None
        self._primary_owner = None
        self._root_organization_policies = None
        self._stripe_billing_history = None
        self._user_organization = None
        self._withdraw_history_organizations = None
        self._workloads = None
        self._workspaces = None
        self.discriminator = None

        if billing_histories is not None:
            self.billing_histories = billing_histories
        if cluster_quotas is not None:
            self.cluster_quotas = cluster_quotas
        if credit_earn_histories is not None:
            self.credit_earn_histories = credit_earn_histories
        if default_storage is not None:
            self.default_storage = default_storage
        if default_volume is not None:
            self.default_volume = default_volume
        if groups is not None:
            self.groups = groups
        if kernel_clusters is not None:
            self.kernel_clusters = kernel_clusters
        if model_services is not None:
            self.model_services = model_services
        if organization_credentials is not None:
            self.organization_credentials = organization_credentials
        if organization_kernel_clusters is not None:
            self.organization_kernel_clusters = organization_kernel_clusters
        if organization_policies is not None:
            self.organization_policies = organization_policies
        if owner is not None:
            self.owner = owner
        if pipeline_step_types is not None:
            self.pipeline_step_types = pipeline_step_types
        if pricing_plan is not None:
            self.pricing_plan = pricing_plan
        if primary_owner is not None:
            self.primary_owner = primary_owner
        if root_organization_policies is not None:
            self.root_organization_policies = root_organization_policies
        if stripe_billing_history is not None:
            self.stripe_billing_history = stripe_billing_history
        if user_organization is not None:
            self.user_organization = user_organization
        if withdraw_history_organizations is not None:
            self.withdraw_history_organizations = withdraw_history_organizations
        if workloads is not None:
            self.workloads = workloads
        if workspaces is not None:
            self.workspaces = workspaces

    @property
    def billing_histories(self):
        """Gets the billing_histories of this OrmOrganizationEdges.  # noqa: E501


        :return: The billing_histories of this OrmOrganizationEdges.  # noqa: E501
        :rtype: list[OrmBillingHistory]
        """
        return self._billing_histories

    @billing_histories.setter
    def billing_histories(self, billing_histories):
        """Sets the billing_histories of this OrmOrganizationEdges.


        :param billing_histories: The billing_histories of this OrmOrganizationEdges.  # noqa: E501
        :type billing_histories: list[OrmBillingHistory]
        """

        self._billing_histories = billing_histories

    @property
    def cluster_quotas(self):
        """Gets the cluster_quotas of this OrmOrganizationEdges.  # noqa: E501


        :return: The cluster_quotas of this OrmOrganizationEdges.  # noqa: E501
        :rtype: list[OrmClusterQuota]
        """
        return self._cluster_quotas

    @cluster_quotas.setter
    def cluster_quotas(self, cluster_quotas):
        """Sets the cluster_quotas of this OrmOrganizationEdges.


        :param cluster_quotas: The cluster_quotas of this OrmOrganizationEdges.  # noqa: E501
        :type cluster_quotas: list[OrmClusterQuota]
        """

        self._cluster_quotas = cluster_quotas

    @property
    def credit_earn_histories(self):
        """Gets the credit_earn_histories of this OrmOrganizationEdges.  # noqa: E501


        :return: The credit_earn_histories of this OrmOrganizationEdges.  # noqa: E501
        :rtype: list[OrmCreditEarnHistory]
        """
        return self._credit_earn_histories

    @credit_earn_histories.setter
    def credit_earn_histories(self, credit_earn_histories):
        """Sets the credit_earn_histories of this OrmOrganizationEdges.


        :param credit_earn_histories: The credit_earn_histories of this OrmOrganizationEdges.  # noqa: E501
        :type credit_earn_histories: list[OrmCreditEarnHistory]
        """

        self._credit_earn_histories = credit_earn_histories

    @property
    def default_storage(self):
        """Gets the default_storage of this OrmOrganizationEdges.  # noqa: E501


        :return: The default_storage of this OrmOrganizationEdges.  # noqa: E501
        :rtype: OrmStorage
        """
        return self._default_storage

    @default_storage.setter
    def default_storage(self, default_storage):
        """Sets the default_storage of this OrmOrganizationEdges.


        :param default_storage: The default_storage of this OrmOrganizationEdges.  # noqa: E501
        :type default_storage: OrmStorage
        """

        self._default_storage = default_storage

    @property
    def default_volume(self):
        """Gets the default_volume of this OrmOrganizationEdges.  # noqa: E501


        :return: The default_volume of this OrmOrganizationEdges.  # noqa: E501
        :rtype: OrmVolume
        """
        return self._default_volume

    @default_volume.setter
    def default_volume(self, default_volume):
        """Sets the default_volume of this OrmOrganizationEdges.


        :param default_volume: The default_volume of this OrmOrganizationEdges.  # noqa: E501
        :type default_volume: OrmVolume
        """

        self._default_volume = default_volume

    @property
    def groups(self):
        """Gets the groups of this OrmOrganizationEdges.  # noqa: E501


        :return: The groups of this OrmOrganizationEdges.  # noqa: E501
        :rtype: list[OrmGroup]
        """
        return self._groups

    @groups.setter
    def groups(self, groups):
        """Sets the groups of this OrmOrganizationEdges.


        :param groups: The groups of this OrmOrganizationEdges.  # noqa: E501
        :type groups: list[OrmGroup]
        """

        self._groups = groups

    @property
    def kernel_clusters(self):
        """Gets the kernel_clusters of this OrmOrganizationEdges.  # noqa: E501


        :return: The kernel_clusters of this OrmOrganizationEdges.  # noqa: E501
        :rtype: list[OrmKernelCluster]
        """
        return self._kernel_clusters

    @kernel_clusters.setter
    def kernel_clusters(self, kernel_clusters):
        """Sets the kernel_clusters of this OrmOrganizationEdges.


        :param kernel_clusters: The kernel_clusters of this OrmOrganizationEdges.  # noqa: E501
        :type kernel_clusters: list[OrmKernelCluster]
        """

        self._kernel_clusters = kernel_clusters

    @property
    def model_services(self):
        """Gets the model_services of this OrmOrganizationEdges.  # noqa: E501


        :return: The model_services of this OrmOrganizationEdges.  # noqa: E501
        :rtype: list[OrmModelService]
        """
        return self._model_services

    @model_services.setter
    def model_services(self, model_services):
        """Sets the model_services of this OrmOrganizationEdges.


        :param model_services: The model_services of this OrmOrganizationEdges.  # noqa: E501
        :type model_services: list[OrmModelService]
        """

        self._model_services = model_services

    @property
    def organization_credentials(self):
        """Gets the organization_credentials of this OrmOrganizationEdges.  # noqa: E501


        :return: The organization_credentials of this OrmOrganizationEdges.  # noqa: E501
        :rtype: list[OrmOrganizationCredentials]
        """
        return self._organization_credentials

    @organization_credentials.setter
    def organization_credentials(self, organization_credentials):
        """Sets the organization_credentials of this OrmOrganizationEdges.


        :param organization_credentials: The organization_credentials of this OrmOrganizationEdges.  # noqa: E501
        :type organization_credentials: list[OrmOrganizationCredentials]
        """

        self._organization_credentials = organization_credentials

    @property
    def organization_kernel_clusters(self):
        """Gets the organization_kernel_clusters of this OrmOrganizationEdges.  # noqa: E501


        :return: The organization_kernel_clusters of this OrmOrganizationEdges.  # noqa: E501
        :rtype: list[OrmOrganizationKernelCluster]
        """
        return self._organization_kernel_clusters

    @organization_kernel_clusters.setter
    def organization_kernel_clusters(self, organization_kernel_clusters):
        """Sets the organization_kernel_clusters of this OrmOrganizationEdges.


        :param organization_kernel_clusters: The organization_kernel_clusters of this OrmOrganizationEdges.  # noqa: E501
        :type organization_kernel_clusters: list[OrmOrganizationKernelCluster]
        """

        self._organization_kernel_clusters = organization_kernel_clusters

    @property
    def organization_policies(self):
        """Gets the organization_policies of this OrmOrganizationEdges.  # noqa: E501


        :return: The organization_policies of this OrmOrganizationEdges.  # noqa: E501
        :rtype: list[OrmAccessControlPolicy]
        """
        return self._organization_policies

    @organization_policies.setter
    def organization_policies(self, organization_policies):
        """Sets the organization_policies of this OrmOrganizationEdges.


        :param organization_policies: The organization_policies of this OrmOrganizationEdges.  # noqa: E501
        :type organization_policies: list[OrmAccessControlPolicy]
        """

        self._organization_policies = organization_policies

    @property
    def owner(self):
        """Gets the owner of this OrmOrganizationEdges.  # noqa: E501


        :return: The owner of this OrmOrganizationEdges.  # noqa: E501
        :rtype: OrmUser
        """
        return self._owner

    @owner.setter
    def owner(self, owner):
        """Sets the owner of this OrmOrganizationEdges.


        :param owner: The owner of this OrmOrganizationEdges.  # noqa: E501
        :type owner: OrmUser
        """

        self._owner = owner

    @property
    def pipeline_step_types(self):
        """Gets the pipeline_step_types of this OrmOrganizationEdges.  # noqa: E501


        :return: The pipeline_step_types of this OrmOrganizationEdges.  # noqa: E501
        :rtype: list[OrmPipelineStepType]
        """
        return self._pipeline_step_types

    @pipeline_step_types.setter
    def pipeline_step_types(self, pipeline_step_types):
        """Sets the pipeline_step_types of this OrmOrganizationEdges.


        :param pipeline_step_types: The pipeline_step_types of this OrmOrganizationEdges.  # noqa: E501
        :type pipeline_step_types: list[OrmPipelineStepType]
        """

        self._pipeline_step_types = pipeline_step_types

    @property
    def pricing_plan(self):
        """Gets the pricing_plan of this OrmOrganizationEdges.  # noqa: E501


        :return: The pricing_plan of this OrmOrganizationEdges.  # noqa: E501
        :rtype: OrmPricingPlan
        """
        return self._pricing_plan

    @pricing_plan.setter
    def pricing_plan(self, pricing_plan):
        """Sets the pricing_plan of this OrmOrganizationEdges.


        :param pricing_plan: The pricing_plan of this OrmOrganizationEdges.  # noqa: E501
        :type pricing_plan: OrmPricingPlan
        """

        self._pricing_plan = pricing_plan

    @property
    def primary_owner(self):
        """Gets the primary_owner of this OrmOrganizationEdges.  # noqa: E501


        :return: The primary_owner of this OrmOrganizationEdges.  # noqa: E501
        :rtype: OrmUser
        """
        return self._primary_owner

    @primary_owner.setter
    def primary_owner(self, primary_owner):
        """Sets the primary_owner of this OrmOrganizationEdges.


        :param primary_owner: The primary_owner of this OrmOrganizationEdges.  # noqa: E501
        :type primary_owner: OrmUser
        """

        self._primary_owner = primary_owner

    @property
    def root_organization_policies(self):
        """Gets the root_organization_policies of this OrmOrganizationEdges.  # noqa: E501


        :return: The root_organization_policies of this OrmOrganizationEdges.  # noqa: E501
        :rtype: list[OrmAccessControlPolicy]
        """
        return self._root_organization_policies

    @root_organization_policies.setter
    def root_organization_policies(self, root_organization_policies):
        """Sets the root_organization_policies of this OrmOrganizationEdges.


        :param root_organization_policies: The root_organization_policies of this OrmOrganizationEdges.  # noqa: E501
        :type root_organization_policies: list[OrmAccessControlPolicy]
        """

        self._root_organization_policies = root_organization_policies

    @property
    def stripe_billing_history(self):
        """Gets the stripe_billing_history of this OrmOrganizationEdges.  # noqa: E501


        :return: The stripe_billing_history of this OrmOrganizationEdges.  # noqa: E501
        :rtype: list[OrmStripeBillingHistory]
        """
        return self._stripe_billing_history

    @stripe_billing_history.setter
    def stripe_billing_history(self, stripe_billing_history):
        """Sets the stripe_billing_history of this OrmOrganizationEdges.


        :param stripe_billing_history: The stripe_billing_history of this OrmOrganizationEdges.  # noqa: E501
        :type stripe_billing_history: list[OrmStripeBillingHistory]
        """

        self._stripe_billing_history = stripe_billing_history

    @property
    def user_organization(self):
        """Gets the user_organization of this OrmOrganizationEdges.  # noqa: E501


        :return: The user_organization of this OrmOrganizationEdges.  # noqa: E501
        :rtype: list[OrmUserOrganization]
        """
        return self._user_organization

    @user_organization.setter
    def user_organization(self, user_organization):
        """Sets the user_organization of this OrmOrganizationEdges.


        :param user_organization: The user_organization of this OrmOrganizationEdges.  # noqa: E501
        :type user_organization: list[OrmUserOrganization]
        """

        self._user_organization = user_organization

    @property
    def withdraw_history_organizations(self):
        """Gets the withdraw_history_organizations of this OrmOrganizationEdges.  # noqa: E501


        :return: The withdraw_history_organizations of this OrmOrganizationEdges.  # noqa: E501
        :rtype: list[OrmWithdrawHistoryOrganization]
        """
        return self._withdraw_history_organizations

    @withdraw_history_organizations.setter
    def withdraw_history_organizations(self, withdraw_history_organizations):
        """Sets the withdraw_history_organizations of this OrmOrganizationEdges.


        :param withdraw_history_organizations: The withdraw_history_organizations of this OrmOrganizationEdges.  # noqa: E501
        :type withdraw_history_organizations: list[OrmWithdrawHistoryOrganization]
        """

        self._withdraw_history_organizations = withdraw_history_organizations

    @property
    def workloads(self):
        """Gets the workloads of this OrmOrganizationEdges.  # noqa: E501


        :return: The workloads of this OrmOrganizationEdges.  # noqa: E501
        :rtype: list[OrmWorkload]
        """
        return self._workloads

    @workloads.setter
    def workloads(self, workloads):
        """Sets the workloads of this OrmOrganizationEdges.


        :param workloads: The workloads of this OrmOrganizationEdges.  # noqa: E501
        :type workloads: list[OrmWorkload]
        """

        self._workloads = workloads

    @property
    def workspaces(self):
        """Gets the workspaces of this OrmOrganizationEdges.  # noqa: E501


        :return: The workspaces of this OrmOrganizationEdges.  # noqa: E501
        :rtype: list[OrmWorkspace]
        """
        return self._workspaces

    @workspaces.setter
    def workspaces(self, workspaces):
        """Sets the workspaces of this OrmOrganizationEdges.


        :param workspaces: The workspaces of this OrmOrganizationEdges.  # noqa: E501
        :type workspaces: list[OrmWorkspace]
        """

        self._workspaces = workspaces

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, OrmOrganizationEdges):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, OrmOrganizationEdges):
            return True

        return self.to_dict() != other.to_dict()
